from io import BytesIO

import pytest

from translate.storage.html import htmlfile


def parse_html(src: str) -> htmlfile:
    f = BytesIO(src.encode("utf-8"))
    f.name = "test.html"
    return htmlfile(f)


def test_html_context_basic() -> None:
    """Test that basic context is extracted correctly from HTML."""
    store = parse_html('<div data-translate-context="greeting">Hello world</div>')
    units = store.getunits()
    assert any(
        u.getcontext() == "greeting" and u.source == "Hello world" for u in units
    )


def test_html_context_attribute() -> None:
    """Test that attribute is added to context."""
    store = parse_html(
        '<p data-translate-context="intro" title="Hello world">Welcome!</p>'
    )
    units = store.getunits()
    unit0 = units[0]
    unit1 = units[1]
    assert unit0.getcontext() == "intro[title]"
    assert unit0.source == "Hello world"
    assert unit1.getcontext() == "intro"
    assert unit1.source == "Welcome!"


def test_html_context_attribute_with_id() -> None:
    """Test that attributes get context hints from element IDs when duplicated."""
    src = '<p id="a" title="Hello">World</p><p id="b" title="Hello">Universe</p>'
    store = parse_html(src)
    title_units = [u for u in store.getunits() if u.source == "Hello"]
    assert len(title_units) == 2
    contexts = {u.getcontext() for u in title_units}
    assert contexts == {"test.html:a[title]", "test.html:b[title]"}


def test_html_context_same_source_different_contexts() -> None:
    """Test that the same source with different contexts is differentiated."""
    store = parse_html(
        '<p data-translate-context="one">Hello</p><p data-translate-context="two">Hello</p>'
    )
    hello_units = [u for u in store.getunits() if u.source == "Hello"]
    contexts = sorted({u.getcontext() for u in hello_units})
    assert contexts == ["one", "two"]
    # Ensure ids differ
    assert len({u.getid() for u in hello_units}) == 2


def test_html_context_nested_outer_wins() -> None:
    """
    Test that outer context applies to outer unit.

    This test uses nested translatable elements (p inside div)
    so that inner becomes its own unit.
    """
    src = (
        '<div data-translate-context="outer">Start '
        '<p data-translate-context="inner">Inner</p> End</div>'
    )
    store = parse_html(src)
    inner = [u for u in store.getunits() if u.source == "Inner"]
    outer = [
        u
        for u in store.getunits()
        if "Start" in u.source or "End" in u.source or u.source.startswith("Start")
    ]
    # Inner unit should have inner context
    assert inner
    assert inner[0].getcontext() == "inner"
    # At least one outer fragment should carry outer context
    assert any(u.getcontext() == "outer" for u in outer)


def test_html_context_absent() -> None:
    """Test that absence of context is handled correctly."""
    store = parse_html("<p>No context here</p>")
    try:
        unit = next(u for u in store.getunits() if u.source == "No context here")
    except StopIteration:
        pytest.fail("Unit with 'No context here' source was not found!")
    assert unit.getcontext() == ""


def test_html_context_id_overridden_by_explicit() -> None:
    """Test that data-translate-context overrides id fallback."""
    store = parse_html('<p id="greeting" data-translate-context="ctx">Hello</p>')
    units = [u for u in store.getunits() if u.source == "Hello"]
    assert units
    assert units[0].getcontext() == "ctx"


def test_html_context_id_not_used_when_no_duplicates() -> None:
    """Test that ID is not used when there are no duplicate sources."""
    src = '<p id="a">Hello</p><p id="b">World</p>'
    store = parse_html(src)
    try:
        hello_unit = next(u for u in store.getunits() if u.source == "Hello")
    except StopIteration:
        pytest.fail("Unit with 'Hello' source was not found!")
    try:
        world_unit = next(u for u in store.getunits() if u.source == "World")
    except StopIteration:
        pytest.fail("Unit with 'World' source was not found!")
    assert hello_unit.getcontext() == ""
    assert world_unit.getcontext() == ""


def test_html_context_id_not_used_when_data_translate_context_identical() -> None:
    """Test that identical data-translate-context has the same context."""
    store = parse_html(
        '<p data-translate-context="greet">Hello</p><div id="lala"><span data-translate-context="greet">Hello</span></div>'
    )
    hello_units = [u for u in store.getunits() if u.source == "Hello"]
    assert len(hello_units) == 1
    assert len(hello_units[0].getlocations()) == 2
    assert hello_units[0].getcontext() == "greet"


def test_html_context_disambiguates_duplicates_with_id() -> None:
    """Test that ID is used to disambiguate when the same source appears multiple times."""
    src = '<p id="a">Hello</p><p id="b">Hello</p>'
    store = parse_html(src)
    hello_units = [u for u in store.getunits() if u.source == "Hello"]
    assert len(hello_units) == 2
    contexts = {u.getcontext() for u in hello_units}
    # Both should be distinct contexts on the basis of their IDs
    assert contexts == {"test.html:a", "test.html:b"}


def test_html_context_disambiguates_duplicates_with_ancestor_id() -> None:
    """Test that when identical sources are under different ancestor IDs, ancestor path hints are applied."""
    src = (
        '<div id="section_a"><p>Hello</p></div>\n<div id="section_b"><p>Hello</p></div>'
    )
    store = parse_html(src)
    units = [u for u in store.getunits() if u.source == "Hello"]
    assert len(units) == 2
    ctx1 = units[0].getcontext()
    ctx2 = units[1].getcontext()
    assert ctx1 == "test.html+section_a.p:0-20"
    assert ctx2 == "test.html+section_b.p:0-20"



import os
import time
from io import BytesIO

from translate.storage import po, poheader, poxliff


def test_parseheaderstring() -> None:
    """Test for the header parsing function."""
    source = r"""item1: one
item2: two:two
this item must get ignored because there is no colon sign in it
item3: three
"""
    d = poheader.parseheaderstring(source)
    print(type(d))
    assert len(d) == 3
    assert d["item1"] == "one"
    assert d["item2"] == "two:two"
    assert d["item3"] == "three"


def test_update() -> None:
    """Test the update function."""
    # do we really add nothing if add==False ?
    d = poheader.update({}, test="hello")
    assert len(d) == 0
    # do we add if add==True ?
    d = poheader.update({}, add=True, Test="hello")
    assert len(d) == 1
    assert d["Test"] == "hello"
    # do we really update ?
    d = poheader.update({"Test": "hello"}, add=True, Test="World")
    assert len(d) == 1
    assert d["Test"] == "World"
    # does key rewrite work ?
    d = poheader.update({}, add=True, test_me="hello")
    assert d["Test-Me"] == "hello"
    # is the order correct ?
    d = {}
    d["Project-Id-Version"] = "abc"
    d["POT-Creation-Date"] = "now"
    d = poheader.update(d, add=True, Test="hello", Report_Msgid_Bugs_To="bugs@list.org")
    assert list(d.keys()) == [
        "Project-Id-Version",
        "Report-Msgid-Bugs-To",
        "POT-Creation-Date",
        "Test",
    ]


def poparse(posource):
    """Helper that parses po source without requiring files."""
    dummyfile = BytesIO(posource.encode())
    return po.pofile(dummyfile)


def poxliffparse(posource):
    """Helper that parses po source into poxliffFile."""
    poxli = poxliff.PoXliffFile()
    poxli.parse(posource)
    return poxli


def check_po_date(datestring) -> None:
    """
    Check the validity of a PO date.

    The datestring must be in the format: 2007-06-08 10:08+0200
    """
    # We don't include the timezone offset as part of our format,
    # because time.strptime() does not recognize %z
    # The use of %z is deprecated in any case.
    date_format = "%Y-%m-%d %H:%M"

    # Get the timezone offset (last 4 digits):
    tz = datestring[-4:]
    assert isinstance(int(tz), int)

    # Strip the timezone from the string, typically something like "+0200".
    # This is to make the datestring conform to the specified format,
    # we can't add %z to the format.
    datestring = datestring[0:-5]

    # Check that the date can be parsed
    assert type(time.strptime(datestring, date_format)) is time.struct_time


def test_po_dates() -> None:
    pofile = po.pofile()
    headerdict = pofile.makeheaderdict(po_revision_date=True)
    check_po_date(headerdict["POT-Creation-Date"])
    check_po_date(headerdict["PO-Revision-Date"])

    headerdict = pofile.makeheaderdict(
        pot_creation_date=time.localtime(), po_revision_date=time.localtime()
    )
    check_po_date(headerdict["POT-Creation-Date"])
    check_po_date(headerdict["PO-Revision-Date"])


def test_timezones() -> None:
    # The following will only work on Unix because of tzset() and %z
    if "tzset" in time.__dict__:
        os.environ["TZ"] = "Asia/Kabul"
        time.tzset()
        assert time.timezone == -16200
        # Typically "+0430"
        assert poheader.tzstring() == time.strftime("%z")

        os.environ["TZ"] = "Asia/Seoul"
        time.tzset()
        assert time.timezone == -32400
        # Typically "+0900"
        assert poheader.tzstring() == time.strftime("%z")

        os.environ["TZ"] = "Africa/Johannesburg"
        time.tzset()
        assert time.timezone == -7200
        # Typically "+0200"
        assert poheader.tzstring() == time.strftime("%z")

        os.environ["TZ"] = "UTC"
        time.tzset()
        assert time.timezone == 0
        # Typically "+0000"
        assert poheader.tzstring() == time.strftime("%z")


def test_header_blank() -> None:
    """Test header functionality."""

    def compare(pofile) -> None:
        print(pofile)
        assert len(pofile.units) == 1
        header = pofile.header()
        assert header.isheader()
        assert not header.isblank()

        headeritems = pofile.parseheader()
        assert headeritems["Project-Id-Version"] == "PACKAGE VERSION"
        assert headeritems["Report-Msgid-Bugs-To"] == ""
        check_po_date(headeritems["POT-Creation-Date"])
        assert headeritems["PO-Revision-Date"] == "YEAR-MO-DA HO:MI+ZONE"
        assert headeritems["Last-Translator"] == "FULL NAME <EMAIL@ADDRESS>"
        assert headeritems["Language-Team"] == "LANGUAGE <LL@li.org>"
        assert headeritems["MIME-Version"] == "1.0"
        assert headeritems["Content-Type"] == "text/plain; charset=UTF-8"
        assert headeritems["Content-Transfer-Encoding"] == "8bit"
        assert headeritems["Plural-Forms"] == "nplurals=INTEGER; plural=EXPRESSION;"

    posource = r"""# other comment\n
msgid ""
msgstr ""
"Project-Id-Version: PACKAGE VERSION\n"
"Report-Msgid-Bugs-To: \n"
"POT-Creation-Date: 2006-03-08 17:30+0200\n"
"PO-Revision-Date: YEAR-MO-DA HO:MI+ZONE\n"
"Last-Translator: FULL NAME <EMAIL@ADDRESS>\n"
"Language-Team: LANGUAGE <LL@li.org>\n"
"MIME-Version: 1.0\n"
"Content-Type: text/plain; charset=UTF-8\n"
"Content-Transfer-Encoding: 8bit\n"
"Plural-Forms: nplurals=INTEGER; plural=EXPRESSION;\n"
"""
    pofile = poparse(posource)
    compare(pofile)


## TODO: enable this code if PoXliffFile is able to parse a header
##
##    poxliffsource = r'''<?xml version="1.0" encoding="utf-8"?>
##<xliff version="1.1" xmlns="urn:oasis:names:tc:xliff:document:1.1">
##
##<file datatype="po" original="test.po" source-language="en-US"><body><trans-unit approved="no" id="1" restype="x-gettext-domain-header" xml:space="preserve"><source>Project-Id-Version: PACKAGE VERSION
##Report-Msgid-Bugs-To:
##POT-Creation-Date: 2006-03-08 17:30+0200
##PO-Revision-Date: YEAR-MO-DA HO:MI+ZONE
##Last-Translator: FULL NAME <ph id="1">&lt;EMAIL@ADDRESS&gt;</ph>
##Language-Team: LANGUAGE <ph id="2">&lt;LL@li.org&gt;</ph>
##MIME-Version: 1.0
##Content-Type: text/plain; charset=UTF-8
##Content-Transfer-Encoding: 8bit
##Plural-Forms: nplurals=INTEGER; plural=EXPRESSION;
##</source><target>Project-Id-Version: PACKAGE VERSION
##Report-Msgid-Bugs-To:
##POT-Creation-Date: 2006-03-08 17:30+0200
##PO-Revision-Date: YEAR-MO-DA HO:MI+ZONE
##Last-Translator: FULL NAME <ph id="1">&lt;EMAIL@ADDRESS&gt;</ph>
##Language-Team: LANGUAGE <ph id="2">&lt;LL@li.org&gt;</ph>
##MIME-Version: 1.0
##Content-Type: text/plain; charset=UTF-8
##Content-Transfer-Encoding: 8bit
##Plural-Forms: nplurals=INTEGER; plural=EXPRESSION;
##</target><context-group name="po-entry" purpose="information"><context context-type="x-po-trancomment">other comment\n</context></context-group><note from="po-translator">other comment\n</note></trans-unit></body></file></xliff>
##'''
##    pofile = poparse(poxliffsource)
##    compare(pofile)


def test_plural_equation() -> None:
    """
    Test that we work with the equation even is the last semicolon is left out, since gettext
    tools don't seem to mind.
    """
    posource = r"""msgid ""
msgstr ""
"Plural-Forms: nplurals=2; plural=(n != 1)%s\n"
"""
    for colon in ("", ";"):
        pofile = poparse(posource % colon)
        print(pofile)
        assert len(pofile.units) == 1
        header = pofile.units[0]
        assert header.isheader()
        assert not header.isblank()

        pofile.parseheader()
        nplural, plural = pofile.getheaderplural()
        assert nplural == "2"
        assert plural == "(n != 1)"


##    TODO: add the same test for PoXliffFile


def test_plural_equation_across_lines() -> None:
    """Test that we work if the plural equation spans more than one line."""
    posource = r"""msgid ""
msgstr ""
"Plural-Forms:  nplurals=3; plural=(n%10==1 && n%100!=11 ? 0 : n%10>=2 && n%"
"10<=4 && (n%100<10 || n%100>=20) ? 1 : 2);\n"
"""
    pofile = poparse(posource)
    print(pofile)
    assert len(pofile.units) == 1
    header = pofile.units[0]
    assert header.isheader()
    assert not header.isblank()

    pofile.parseheader()
    nplural, plural = pofile.getheaderplural()
    assert nplural == "3"
    assert (
        plural
        == "(n%10==1 && n%100!=11 ? 0 : n%10>=2 && n%10<=4 && (n%100<10 || n%100>=20) ? 1 : 2)"
    )


##    TODO: add the same test for PoXliffFile


def test_updatecontributor() -> None:
    """Test that we can update contributor information in the header comments."""
    posource = r"""msgid ""
msgstr ""
"MIME-Version: 1.0"
"""
    pofile = poparse(posource)
    pofile.updatecontributor("Grasvreter")
    assert "# Grasvreter, 20" in bytes(pofile).decode("utf-8")

    pofile.updatecontributor("Koeivreter", "monster@grasveld.moe")
    assert "# Koeivreter <monster@grasveld.moe>, 20" in bytes(pofile).decode("utf-8")

    pofile.header().addnote("Khaled Hosny <khaledhosny@domain.org>, 2006, 2007, 2008.")
    pofile.updatecontributor("Khaled Hosny", "khaledhosny@domain.org")
    print(bytes(pofile))
    assert (
        f"# Khaled Hosny <khaledhosny@domain.org>, 2006, 2007, 2008, {time.strftime('%Y')}."
        in bytes(pofile).decode("utf-8")
    )


def test_updatecontributor_header() -> None:
    """Test preserving empty lines in comments."""
    posource = r"""# Japanese translation of ibus.
# Copyright (C) 2015-2019 Takao Fujiwara <takao.fujiwara1@gmail.com>
# This file is distributed under the same license as the ibus package.
#
# Translators:
# Takao Fujiwara <takao.fujiwara1@gmail.com>, 2012
msgid ""
msgstr ""
"MIME-Version: 1.0"
"""
    pofile = poparse(posource)

    # Add contributor
    pofile.updatecontributor("Grasvreter")

    # Manually build expected output
    expected = posource.replace("msgid", f"# Grasvreter, {time.strftime('%Y')}.\nmsgid")
    assert bytes(pofile).decode("utf-8") == expected


def test_language() -> None:
    """Test that we can get a language from the relevant headers."""
    posource = r"""msgid ""
msgstr ""
"MIME-Version: 1.0\n"
"""

    pofile = poparse(posource)
    assert pofile.gettargetlanguage() is None

    posource += '"Language-Team: translate-discuss-af@lists.sourceforge.net\\n"\n'
    pofile = poparse(posource)
    assert pofile.gettargetlanguage() == "af"

    posource += '"X-Poedit-Language: German\\n"\n'
    pofile = poparse(posource)
    assert pofile.gettargetlanguage() == "de"

    posource += '"Language: fr_CA\\n"\n'
    pofile = poparse(posource)
    assert pofile.gettargetlanguage() == "fr_CA"


def test_project() -> None:
    """Test that we can get a project from the relevant headers."""
    posource = r"""msgid ""
msgstr ""
"MIME-Version: 1.0\n"
"""

    pofile = poparse(posource)
    assert pofile.getprojectstyle() is None

    posource += '"X-Accelerator-Marker: ~\\n"\n'
    pofile = poparse(posource)
    assert pofile.getprojectstyle() == "openoffice"

    posource += '"Report-Msgid-Bugs-To: http://bugzilla.gnome.org/enter_bug.cgi?product=system-\\n"\n'
    pofile = poparse(posource)
    assert pofile.getprojectstyle() == "gnome"

    posource += '"X-Project-Style: drupal\\n"\n'
    pofile = poparse(posource)
    assert pofile.getprojectstyle() == "drupal"

    pofile.setprojectstyle("kde")
    assert pofile.getprojectstyle() == "kde"

    pofile.setprojectstyle("complete-rubbish")
    assert pofile.getprojectstyle() == "complete-rubbish"



import pytest
from lxml import etree

from translate.misc.xml_helpers import setXMLspace
from translate.storage import xliff
from translate.storage.placeables import StringElem
from translate.storage.placeables.xliff import G, X

from . import test_base


class TestXLIFFUnit(test_base.TestTranslationUnit):
    UnitClass = xliff.xliffunit

    def test_markreview(self) -> None:
        """Tests if we can mark the unit to need review."""
        unit = self.unit
        # We have to explicitly set the target to nothing, otherwise xliff
        # tests will fail.
        # Can we make it default behavior for the UnitClass?
        unit.target = ""

        unit.addnote("Test note 1", origin="translator")
        unit.addnote("Test note 2", origin="translator")
        original_notes = unit.getnotes(origin="translator")

        assert not unit.isreview()
        unit.markreviewneeded()
        assert unit.isreview()
        unit.markreviewneeded(False)
        assert not unit.isreview()
        assert unit.getnotes(origin="translator") == original_notes
        unit.markreviewneeded(explanation="Double check spelling.")
        assert unit.isreview()
        notes = unit.getnotes(origin="translator")
        assert notes.count("Double check spelling.") == 1

    def test_errors(self) -> None:
        """Tests that we can add and retrieve error messages for a unit."""
        unit = self.unit

        assert len(unit.geterrors()) == 0
        unit.adderror(errorname="test1", errortext="Test error message 1.")
        unit.adderror(errorname="test2", errortext="Test error message 2.")
        unit.adderror(errorname="test3", errortext="Test error message 3.")
        assert len(unit.geterrors()) == 3
        assert unit.geterrors()["test1"] == "Test error message 1."
        assert unit.geterrors()["test2"] == "Test error message 2."
        assert unit.geterrors()["test3"] == "Test error message 3."
        unit.adderror(errorname="test1", errortext="New error 1.")
        assert unit.geterrors()["test1"] == "New error 1."

    def test_accepted_control_chars(self) -> None:
        """
        Tests we can assign the accepted control chars.

        Source: :wp:`Valid_characters_in_XML#XML_1.0`
        """
        # Unicode Character 'CHARACTER TABULATION' (U+0009)
        self.unit.target = "Een\t"
        assert self.unit.target == "Een\t"
        # Unicode Character 'LINE FEED (LF)' (U+000A)
        self.unit.target = "Een\n"
        assert self.unit.target == "Een\n"
        # Unicode Character 'CARRIAGE RETURN (CR)' (U+000D)
        self.unit.target = "Een\r"
        assert self.unit.target == "Een\r"

    def test_unaccepted_control_chars(self) -> None:
        """
        Tests we cannot assign the unaccepted control chars without escaping.

        Source: :wp:`Valid_characters_in_XML#XML_1.0`
        """
        self.unit.target = "Een\x00"
        assert self.unit.target == "Een"


class TestXLIFFfile(test_base.TestTranslationStore):
    StoreClass = xliff.xlifffile
    skeleton = """<?xml version="1.0" encoding="utf-8"?>
<xliff version="1.1" xmlns="urn:oasis:names:tc:xliff:document:1.1">
        <file original="doc.txt" source-language="en-US">
                <body>
                        %s
                </body>
        </file>
</xliff>"""

    def test_basic(self) -> None:
        xlifffile = xliff.xlifffile()
        assert xlifffile.units == []
        xlifffile.addsourceunit("Bla")
        assert len(xlifffile.units) == 1
        newfile = xliff.xlifffile.parsestring(bytes(xlifffile))
        print(bytes(xlifffile))
        assert len(newfile.units) == 1
        assert newfile.units[0].source == "Bla"
        assert newfile.findunit("Bla").source == "Bla"
        assert newfile.findunit("dit") is None

    def test_namespace(self) -> None:
        """Check that we handle namespaces other than the default correctly."""
        xlfsource = """<?xml version="1.0" encoding="utf-8"?>
<xliff:xliff version="1.2" xmlns:xliff="urn:oasis:names:tc:xliff:document:1.2">
    <xliff:file original="doc.txt" source-language="en-US">
        <xliff:body>
            <xliff:trans-unit id="1">
                <xliff:source>File 1</xliff:source>
            </xliff:trans-unit>
        </xliff:body>
    </xliff:file>
</xliff:xliff>"""
        xlifffile = xliff.xlifffile.parsestring(xlfsource)
        print(bytes(xlifffile))
        assert xlifffile.units[0].source == "File 1"

    def test_rich_source(self) -> None:
        xlifffile = xliff.xlifffile()
        xliffunit = xlifffile.addsourceunit("")

        # Test 1
        xliffunit.rich_source = [StringElem(["foo", X(id="bar"), "baz"])]
        source_dom_node = xliffunit.getlanguageNode(None, 0)
        x_placeable = source_dom_node[0]

        assert source_dom_node.text == "foo"

        assert x_placeable.tag == "x"
        assert x_placeable.attrib["id"] == "bar"
        assert x_placeable.tail == "baz"

        xliffunit.rich_source[0].print_tree(2)
        print(xliffunit.rich_source)
        assert xliffunit.rich_source == [
            StringElem([StringElem("foo"), X(id="bar"), StringElem("baz")])
        ]

        # Test 2
        xliffunit.rich_source = [
            StringElem(
                ["foo", "baz", G(id="oof", sub=[G(id="zab", sub=["bar", "rab"])])]
            )
        ]
        source_dom_node = xliffunit.getlanguageNode(None, 0)
        g_placeable = source_dom_node[0]
        nested_g_placeable = g_placeable[0]

        assert source_dom_node.text == "foobaz"

        assert g_placeable.tag == "g"
        assert g_placeable.text is None
        assert g_placeable.attrib["id"] == "oof"
        assert g_placeable.tail is None

        assert nested_g_placeable.tag == "g"
        assert nested_g_placeable.text == "barrab"
        assert nested_g_placeable.attrib["id"] == "zab"
        assert nested_g_placeable.tail is None

        rich_source = xliffunit.rich_source
        rich_source[0].print_tree(2)
        assert rich_source == [
            StringElem(["foobaz", G(id="oof", sub=[G(id="zab", sub=["barrab"])])])
        ]

    def test_rich_target(self) -> None:
        xlifffile = xliff.xlifffile()
        xliffunit = xlifffile.addsourceunit("")

        # Test 1
        xliffunit.set_rich_target([StringElem(["foo", X(id="bar"), "baz"])], "fr")
        target_dom_node = xliffunit.getlanguageNode(None, 1)
        x_placeable = target_dom_node[0]

        assert target_dom_node.text == "foo"
        assert x_placeable.tag == "x"
        assert x_placeable.attrib["id"] == "bar"
        assert x_placeable.tail == "baz"

        # Test 2
        xliffunit.target = "test plain target"
        xliffunit.set_rich_target(
            [StringElem(["foo", G(id="eek", sub=[G(id="ook", sub=["bar", "rab"])])])],
            "fr",
        )
        xliffunit.set_rich_target(
            [
                StringElem(
                    ["foo", "baz", G(id="oof", sub=[G(id="zab", sub=["bar", "rab"])])]
                )
            ],
            "fr",
        )
        target_dom_node = xliffunit.getlanguageNode(None, 1)
        g_placeable = target_dom_node[0]
        nested_g_placeable = g_placeable[0]

        assert target_dom_node.text == "foobaz"

        assert g_placeable.tag == "g"
        print(f"g_placeable.text: {g_placeable.text} ({type(g_placeable.text)})")
        assert g_placeable.text is None
        assert g_placeable.attrib["id"] == "oof"
        assert g_placeable.tail is None

        assert nested_g_placeable.tag == "g"
        assert nested_g_placeable.text == "barrab"
        assert nested_g_placeable.attrib["id"] == "zab"
        assert nested_g_placeable.tail is None

        xliffunit.rich_target[0].print_tree(2)
        assert xliffunit.rich_target == [
            StringElem(["foobaz", G(id="oof", sub=[G(id="zab", sub=["barrab"])])])
        ]
        print(bytes(xlifffile).decode())
        assert (
            bytes(xlifffile).decode()
            == """<?xml version="1.0" encoding="UTF-8"?>
<xliff xmlns="urn:oasis:names:tc:xliff:document:1.1" version="1.1">
  <file original="NoName" source-language="en" datatype="plaintext">
    <body>
      <trans-unit xml:space="preserve" id="2" approved="yes">
        <source></source>
        <target state="translated">foobaz<g id="oof"><g id="zab">barrab</g></g></target>
      </trans-unit>
    </body>
  </file>
</xliff>
"""
        )

    def test_rich_source_preserves_whitespace_with_inline_markup(self) -> None:
        xlfsource = """<?xml version="1.0" encoding="utf-8"?>
<xliff version="1.1" xmlns="urn:oasis:names:tc:xliff:document:1.1">
    <file original="doc.txt" source-language="en-US">
        <body>
            <trans-unit id="preserve-space-inline-markup" maxwidth="2000" size-unit="char" approved="yes">
                <source xml:space="preserve">Label A: <mrk mtype="protected">%1</mrk> | Label B: <mrk mtype="protected">%2</mrk> |  Label C: <mrk mtype="protected">%3</mrk></source>
                <target xml:space="preserve" state="final">Translated A:<mrk mtype="protected">%1</mrk> | Translated B: <mrk mtype="protected">%2</mrk> |  Translated C: <mrk mtype="protected">%3</mrk></target>
            </trans-unit>
        </body>
    </file>
</xliff>"""
        xlifffile = xliff.xlifffile.parsestring(xlfsource)
        xliffunit = xlifffile.units[0]

        assert xliffunit.source == "Label A: %1 | Label B: %2 |  Label C: %3"
        assert (
            str(xliffunit.rich_source[0]) == "Label A: %1 | Label B: %2 |  Label C: %3"
        )
        assert (
            xliffunit.target == "Translated A:%1 | Translated B: %2 |  Translated C: %3"
        )
        assert (
            str(xliffunit.rich_target[0])
            == "Translated A:%1 | Translated B: %2 |  Translated C: %3"
        )
        assert (
            bytes(xlifffile)
            .decode()
            .count('|  Label C: <mrk mtype="protected">%3</mrk>')
            == 1
        )
        assert (
            bytes(xlifffile)
            .decode()
            .count('|  Translated C: <mrk mtype="protected">%3</mrk>')
            == 1
        )

    def test_source(self) -> None:
        xlifffile = xliff.xlifffile()
        xliffunit = xlifffile.addsourceunit("Concept")
        xliffunit.source = "Term"
        newfile = xliff.xlifffile.parsestring(bytes(xlifffile))
        print(bytes(xlifffile))
        assert newfile.findunit("Concept") is None
        assert newfile.findunit("Term") is not None

    def test_target(self) -> None:
        xlifffile = xliff.xlifffile()
        xliffunit = xlifffile.addsourceunit("Concept")
        xliffunit.target = "Konsep"
        newfile = xliff.xlifffile.parsestring(bytes(xlifffile))
        print(bytes(xlifffile))
        assert newfile.findunit("Concept").target == "Konsep"

    def test_sourcelanguage(self) -> None:
        xlifffile = xliff.xlifffile(sourcelanguage="xh")
        xmltext = bytes(xlifffile).decode("utf-8")
        print(xmltext)
        assert xmltext.find('source-language="xh"') > 0
        # TODO: test that it also works for new files.

    def test_targetlanguage(self) -> None:
        xlifffile = xliff.xlifffile(sourcelanguage="zu", targetlanguage="af")
        xmltext = bytes(xlifffile).decode("utf-8")
        print(xmltext)
        assert xmltext.find('source-language="zu"') > 0
        assert xmltext.find('target-language="af"') > 0

    def test_targetlanguage_multi(self) -> None:
        xlfsource = """<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE foo [ <!ELEMENT foo ANY > <!ENTITY xxe SYSTEM "file:///etc/passwd" >]>
<xliff version="1.1" xmlns="urn:oasis:names:tc:xliff:document:1.1">
        <file original="doc.txt" source-language="en-US">
        </file>
        <file original="doc.txt" source-language="en-US">
        </file>
</xliff>"""
        xlifffile = xliff.xlifffile.parsestring(xlfsource)
        xlifffile.settargetlanguage("cs")
        xlifffile.setsourcelanguage("de")
        xmltext = bytes(xlifffile).decode()
        print(xmltext)
        assert xmltext.count('source-language="de"') == 2
        assert xmltext.count('target-language="cs"') == 2

    def test_notes(self) -> None:
        xlifffile = xliff.xlifffile()
        unit = xlifffile.addsourceunit("Concept")
        # We don't want to add unnecessary notes
        assert "note" not in bytes(xlifffile).decode("utf-8")
        unit.addnote(None)
        assert "note" not in bytes(xlifffile).decode("utf-8")
        unit.addnote("")
        assert "note" not in bytes(xlifffile).decode("utf-8")

        unit.addnote("Please buy bread")
        assert unit.getnotes() == "Please buy bread"
        notenodes = unit.xmlelement.findall(f".//{unit.namespaced('note')}")
        assert len(notenodes) == 1

        unit.addnote("Please buy milk", origin="Mom")
        notenodes = unit.xmlelement.findall(f".//{unit.namespaced('note')}")
        assert len(notenodes) == 2
        assert "from" not in notenodes[0].attrib
        assert notenodes[1].get("from") == "Mom"
        assert unit.getnotes(origin="Mom") == "Please buy milk"

        unit.addnote("Don't forget the beer", origin="Dad")
        notenodes = unit.xmlelement.findall(f".//{unit.namespaced('note')}")
        assert len(notenodes) == 3
        assert notenodes[1].get("from") == "Mom"
        assert notenodes[2].get("from") == "Dad"
        assert unit.getnotes(origin="Dad") == "Don't forget the beer"

        assert (
            unit.getnotes(origin="Bob")
            != "Please buy bread\nPlease buy milk\nDon't forget the beer"
        )
        assert notenodes[2].get("from") != "Mom"
        assert "from" not in notenodes[0].attrib
        assert (
            unit.getnotes()
            == "Please buy bread\nPlease buy milk\nDon't forget the beer"
        )
        assert unit.correctorigin(notenodes[2], "ad")
        assert not unit.correctorigin(notenodes[2], "om")

    def test_alttrans(self) -> None:
        """Test xliff <alt-trans> accessors."""
        xlifffile = xliff.xlifffile()
        unit = xlifffile.addsourceunit("Testing")

        unit.addalttrans("ginmi")
        unit.addalttrans("shikenki")
        alternatives = unit.getalttrans()
        assert alternatives[0].source == "Testing"
        assert alternatives[0].target == "ginmi"
        assert alternatives[1].target == "shikenki"

        assert not unit.target

        unit.addalttrans("Tasting", origin="bob", lang="eng")
        alternatives = unit.getalttrans()
        assert alternatives[2].target == "Tasting"

        alternatives = unit.getalttrans(origin="bob")
        assert alternatives[0].target == "Tasting"

        unit.delalttrans(alternatives[0])
        assert len(unit.getalttrans(origin="bob")) == 0
        alternatives = unit.getalttrans()
        assert len(alternatives) == 2
        assert alternatives[0].target == "ginmi"
        assert alternatives[1].target == "shikenki"

        # clean up:
        alternatives = unit.getalttrans()
        for alt in alternatives:
            unit.delalttrans(alt)
        unit.addalttrans("targetx", sourcetxt="sourcex")
        # test that the source node is before the target node:
        alt = unit.getalttrans()[0]
        altformat = etree.tostring(alt.xmlelement, encoding="unicode")
        print(altformat)
        assert altformat.find("<source") < altformat.find("<target")

        # test that a new target is still before alt-trans (bug 1098)
        unit.target = "newester target"
        unitformat = str(unit)
        print(unitformat)
        assert (
            unitformat.find("<source")
            < unitformat.find("<target")
            < unitformat.find("<alt-trans")
        )

    def test_fuzzy(self) -> None:
        xlifffile = xliff.xlifffile()
        unit = xlifffile.addsourceunit("Concept")
        unit.markfuzzy()
        assert not unit.isfuzzy()  # untranslated
        unit.target = "Konsep"
        assert not unit.isfuzzy()  # translated
        unit.markfuzzy()
        assert unit.isfuzzy()
        unit.markfuzzy(False)
        assert not unit.isfuzzy()
        unit.markfuzzy(True)
        assert unit.isfuzzy()

        # If there is no target, we can't really indicate fuzzyness, so we set
        # approved to "no". If we want isfuzzy() to reflect that, the line can
        # be uncommented
        unit.target = None
        assert unit.target is None
        print(unit)
        unit.markfuzzy(True)
        assert 'approved="no"' in str(unit)
        # assert unit.isfuzzy()

    def test_xml_space(self) -> None:
        """Test for the correct handling of xml:space attributes."""
        xlfsource = self.skeleton % (
            """<trans-unit id="1" xml:space="preserve">
                   <source> File  1 </source>
               </trans-unit>"""
        )
        xlifffile = xliff.xlifffile.parsestring(xlfsource)
        assert xlifffile.units[0].source == " File  1 "
        root_node = xlifffile.document.getroot()
        setXMLspace(root_node, "preserve")
        assert xlifffile.units[0].source == " File  1 "
        setXMLspace(root_node, "default")
        assert xlifffile.units[0].source == " File  1 "

        xlfsource = self.skeleton % (
            """<trans-unit id="1" xml:space="default">
                   <source> File  1 </source>
               </trans-unit>"""
        )
        xlifffile = xliff.xlifffile.parsestring(xlfsource)
        assert xlifffile.units[0].source == "File 1"
        root_node = xlifffile.document.getroot()
        setXMLspace(root_node, "preserve")
        assert xlifffile.units[0].source == "File 1"
        setXMLspace(root_node, "default")
        assert xlifffile.units[0].source == "File 1"

        xlfsource = self.skeleton % (
            """<trans-unit id="1">
                   <source> File  1 </source>
               </trans-unit>"""
        )
        # we currently always normalize as default behaviour for xliff
        xlifffile = xliff.xlifffile.parsestring(xlfsource)
        assert xlifffile.units[0].source == "File 1"
        root_node = xlifffile.document.getroot()
        setXMLspace(root_node, "preserve")
        assert xlifffile.units[0].source == "File 1"
        setXMLspace(root_node, "default")
        assert xlifffile.units[0].source == "File 1"

        xlfsource = self.skeleton % (
            """<trans-unit id="1">
                   <source> File  1
</source>
               </trans-unit>"""
        )
        # we currently always normalize as default behaviour for xliff
        xlifffile = xliff.xlifffile.parsestring(xlfsource)
        assert xlifffile.units[0].source == "File 1"
        root_node = xlifffile.document.getroot()
        setXMLspace(root_node, "preserve")
        assert xlifffile.units[0].source == "File 1"
        setXMLspace(root_node, "default")
        assert xlifffile.units[0].source == "File 1"

    def test_parsing(self) -> None:
        xlfsource = (
            self.skeleton
            % """<trans-unit id="1" xml:space="preserve">
                     <source>File</source>
                     <target/>
                 </trans-unit>"""
        )
        xlifffile = xliff.xlifffile.parsestring(xlfsource)
        assert xlifffile.units[0].istranslatable()

        xlfsource = (
            self.skeleton
            % """<trans-unit id="1" xml:space="preserve" translate="no">
                     <source>File</source>
                     <target/>
                 </trans-unit>"""
        )
        xlifffile = xliff.xlifffile.parsestring(xlfsource)
        assert not xlifffile.units[0].istranslatable()

        xlfsource = (
            self.skeleton
            % """<trans-unit id="1" xml:space="preserve" translate="yes">
                     <source>File</source>
                     <target/>
                 </trans-unit>"""
        )
        xlifffile = xliff.xlifffile.parsestring(xlfsource)
        assert xlifffile.units[0].istranslatable()

    def test_marktranslatable(self) -> None:
        """Test marking units as translatable or untranslatable."""
        xlfsource = (
            self.skeleton
            % """<trans-unit id="1" xml:space="preserve">
                     <source>File</source>
                     <target/>
                 </trans-unit>"""
        )
        xlifffile = xliff.xlifffile.parsestring(xlfsource)
        unit = xlifffile.units[0]

        # Test marking as untranslatable
        assert unit.istranslatable()
        unit.marktranslatable(False)
        assert not unit.istranslatable()
        assert unit.xmlelement.get("translate") == "no"

        # Test marking as translatable
        unit.marktranslatable(True)
        assert unit.istranslatable()
        assert unit.xmlelement.get("translate") == "yes"

        # Test that marking as translatable again when already translatable sets to "yes"
        unit.marktranslatable(True)
        assert unit.istranslatable()
        assert unit.xmlelement.get("translate") == "yes"

        # Test that marking as untranslatable when already untranslatable doesn't change it
        unit.marktranslatable(False)
        unit.marktranslatable(False)
        assert not unit.istranslatable()
        assert unit.xmlelement.get("translate") == "no"

    def test_entities(self) -> None:
        xlfsource = """<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE foo [ <!ELEMENT foo ANY > <!ENTITY xxe SYSTEM "file:///etc/passwd" >]>
<xliff version="1.1" xmlns="urn:oasis:names:tc:xliff:document:1.1">
        <file original="doc.txt" source-language="en-US">
                <body>
                    <trans-unit id="1" xml:space="preserve" translate="yes">
                        <source>&xxe;</source>
                        <target/>
                    </trans-unit>
                    <trans-unit id="2" xml:space="preserve" translate="yes">
                        <source>&amp;</source>
                        <target/>
                    </trans-unit>
                </body>
        </file>
</xliff>"""
        xlifffile = xliff.xlifffile.parsestring(xlfsource)
        assert xlifffile.units[0].istranslatable()
        assert xlifffile.units[0].source == ""
        assert xlifffile.units[1].istranslatable()
        assert xlifffile.units[1].source == "&"

    def test_multiple_filenodes(self) -> None:
        xlfsource = """<?xml version="1.0" encoding="utf-8"?>
<xliff version="1.1" xmlns="urn:oasis:names:tc:xliff:document:1.1">
  <file original="file0" source-language="en" datatype="plaintext">
    <body>
      <trans-unit id="hello" approved="yes">
        <source>Hello</source>
      </trans-unit>
    </body>
  </file>
  <file original="file1" source-language="en" datatype="plaintext">
    <body>
      <trans-unit id="world" approved="yes">
        <source>World</source>
      </trans-unit>
    </body>
  </file>
</xliff>"""
        xfile = xliff.xlifffile.parsestring(xlfsource)
        assert len(xfile.units) == 2
        assert xfile.units[0].getid() == "file0\x04hello"
        assert xfile.units[1].getid() == "file1\x04world"
        xunit = xliff.xlifffile.UnitClass(source="goodbye")
        xunit.setid("file2\x04goodbye")
        xfile.addunit(xunit)
        assert xfile.units[2].getid() == "file2\x04goodbye"
        # if there is no file set it will use the active context
        xunit = xliff.xlifffile.UnitClass(source="lastfile")
        xunit.setid("lastfile")
        xfile.addunit(xunit)
        assert xfile.units[3].getid() == "file2\x04lastfile"
        newxfile = xliff.xlifffile()
        newxfile.addunit(xfile.units[0])
        newxfile.addunit(xfile.units[1])
        assert newxfile.units[0].getid() == "file0\x04hello"
        assert newxfile.units[1].getid() == "file1\x04world"
        assert newxfile.getfilenode("file0") is not None
        assert newxfile.getfilenode("file1") is not None
        assert not newxfile.getfilenode("foo")

    def test_preserve_groups_when_adding_units(self) -> None:
        """Test that groups are preserved when adding units from one store to another."""
        source_xliff = b"""<?xml version="1.0" encoding="utf-8"?>
<xliff version="1.1" xmlns="urn:oasis:names:tc:xliff:document:1.1">
    <file original="file1.txt" source-language="en" datatype="plaintext">
        <body>
            <group id="group1" restype="x-gettext-domain">
                <trans-unit id="hello">
                    <source>Hello</source>
                </trans-unit>
            </group>
            <group id="group2">
                <trans-unit id="world">
                    <source>World</source>
                </trans-unit>
            </group>
        </body>
    </file>
</xliff>"""

        expected_output = b"""<?xml version="1.0" encoding="UTF-8"?>
<xliff xmlns="urn:oasis:names:tc:xliff:document:1.1" version="1.1">
  <file original="file1.txt" source-language="en" datatype="plaintext" target-language="fr">
    <body>
      <group id="group1" restype="x-gettext-domain">
        <trans-unit id="hello">
          <source>Hello</source>
        </trans-unit>
      </group>
      <group id="group2">
        <trans-unit id="world">
          <source>World</source>
        </trans-unit>
      </group>
    </body>
  </file>
</xliff>
"""

        source = xliff.xlifffile.parsestring(source_xliff)
        assert len(source.units) == 2

        # Add units to a new translation store
        translation = xliff.xlifffile()
        translation.sourcelanguage = "en"
        translation.targetlanguage = "fr"

        for unit in source.units:
            translation.addunit(unit)

        # Verify the structure matches expected
        serialized = bytes(translation)
        assert serialized == expected_output

        # Also verify by parsing and checking structure
        translation_reloaded = xliff.xlifffile.parsestring(serialized)
        assert len(translation_reloaded.units) == 2
        for unit in translation_reloaded.units:
            parent = unit.xmlelement.getparent()
            assert etree.QName(parent).localname == "group"

    def test_preserve_multiple_files_and_groups(self) -> None:
        """Test that both files and groups are preserved."""
        source_xliff = b"""<?xml version="1.0" encoding="utf-8"?>
<xliff version="1.1" xmlns="urn:oasis:names:tc:xliff:document:1.1">
    <file original="file1.txt" source-language="en" datatype="plaintext">
        <body>
            <group id="group1">
                <trans-unit id="hello">
                    <source>Hello</source>
                </trans-unit>
            </group>
        </body>
    </file>
    <file original="file2.txt" source-language="en" datatype="plaintext">
        <body>
            <group id="group2">
                <trans-unit id="foo">
                    <source>Foo</source>
                </trans-unit>
            </group>
        </body>
    </file>
</xliff>"""

        source = xliff.xlifffile.parsestring(source_xliff)
        assert len(source.units) == 2

        # Add units to a new translation store
        translation = xliff.xlifffile()
        translation.sourcelanguage = "en"
        translation.targetlanguage = "fr"

        for unit in source.units:
            translation.addunit(unit)

        # Verify both files and groups are preserved
        assert translation.getfilenames() == ["file1.txt", "file2.txt"]
        serialized = bytes(translation)
        translation_reloaded = xliff.xlifffile.parsestring(serialized)

        # Verify structure matches original
        assert translation_reloaded.getfilenames() == ["file1.txt", "file2.txt"]
        assert len(translation_reloaded.units) == 2

        # Check that each unit is in the correct file and group
        for i, unit in enumerate(translation_reloaded.units):
            parent = unit.xmlelement.getparent()
            assert etree.QName(parent).localname == "group"
            file_node = None
            for ancestor in unit.xmlelement.iterancestors():
                if ancestor.get("original"):
                    file_node = ancestor.get("original")
                    break
            if i == 0:
                assert file_node == "file1.txt"
                assert parent.get("id") == "group1"
            else:
                assert file_node == "file2.txt"
                assert parent.get("id") == "group2"

    def test_add_unit_to_existing_group(self) -> None:
        """Test that new units are added to existing groups when appropriate."""
        # Start with translation that has one group
        translation_xliff = b"""<?xml version="1.0" encoding="utf-8"?>
<xliff version="1.1" xmlns="urn:oasis:names:tc:xliff:document:1.1">
    <file original="file1.txt" source-language="en" datatype="plaintext" target-language="fr">
        <body>
            <group id="group1">
                <trans-unit id="hello">
                    <source>Hello</source>
                    <target>Bonjour</target>
                </trans-unit>
            </group>
        </body>
    </file>
</xliff>"""

        translation = xliff.xlifffile.parsestring(translation_xliff)

        # Add a new unit from a source that has the same group
        source_xliff = b"""<?xml version="1.0" encoding="utf-8"?>
<xliff version="1.1" xmlns="urn:oasis:names:tc:xliff:document:1.1">
    <file original="file1.txt" source-language="en" datatype="plaintext">
        <body>
            <group id="group1">
                <trans-unit id="world">
                    <source>World</source>
                </trans-unit>
            </group>
        </body>
    </file>
</xliff>"""

        source = xliff.xlifffile.parsestring(source_xliff)
        new_unit = source.units[0]

        # Add the new unit
        translation.addunit(new_unit)

        # Verify the new unit is in the same group
        assert len(translation.units) == 2
        serialized = bytes(translation)
        translation_reloaded = xliff.xlifffile.parsestring(serialized)

        # Both units should be in group1
        for unit in translation_reloaded.units:
            parent = unit.xmlelement.getparent()
            assert etree.QName(parent).localname == "group"
            assert parent.get("id") == "group1"

    def test_add_unit_to_different_file(self) -> None:
        """Test adding units with different file than existing ones."""
        # Start with translation that has one file
        translation_xliff = b"""<?xml version="1.0" encoding="utf-8"?>
<xliff version="1.1" xmlns="urn:oasis:names:tc:xliff:document:1.1">
    <file original="file1.txt" source-language="en" datatype="plaintext" target-language="fr">
        <body>
            <trans-unit id="hello">
                <source>Hello</source>
                <target>Bonjour</target>
            </trans-unit>
        </body>
    </file>
</xliff>"""

        translation = xliff.xlifffile.parsestring(translation_xliff)
        assert translation.getfilenames() == ["file1.txt"]

        # Add a new unit from a different file
        source_xliff = b"""<?xml version="1.0" encoding="utf-8"?>
<xliff version="1.1" xmlns="urn:oasis:names:tc:xliff:document:1.1">
    <file original="file2.txt" source-language="en" datatype="plaintext">
        <body>
            <trans-unit id="world">
                <source>World</source>
            </trans-unit>
        </body>
    </file>
</xliff>"""

        source = xliff.xlifffile.parsestring(source_xliff)
        new_unit = source.units[0]

        # Add the new unit
        translation.addunit(new_unit)

        # Verify both files exist
        assert len(translation.units) == 2
        assert set(translation.getfilenames()) == {"file1.txt", "file2.txt"}

        # Verify structure is correct
        serialized = bytes(translation)
        translation_reloaded = xliff.xlifffile.parsestring(serialized)
        assert set(translation_reloaded.getfilenames()) == {"file1.txt", "file2.txt"}

        # Verify each unit is in the correct file
        for unit in translation_reloaded.units:
            unit_id = unit.getid()
            if "hello" in unit_id:
                assert unit_id.startswith("file1.txt")
            elif "world" in unit_id:
                assert unit_id.startswith("file2.txt")

    def test_mixed_groups_and_body(self) -> None:
        """Test files with both grouped and non-grouped units."""
        source_xliff = b"""<?xml version="1.0" encoding="utf-8"?>
<xliff version="1.1" xmlns="urn:oasis:names:tc:xliff:document:1.1">
    <file original="file1.txt" source-language="en" datatype="plaintext">
        <body>
            <group id="group1">
                <trans-unit id="hello">
                    <source>Hello</source>
                </trans-unit>
            </group>
            <trans-unit id="world">
                <source>World</source>
            </trans-unit>
        </body>
    </file>
</xliff>"""

        source = xliff.xlifffile.parsestring(source_xliff)
        translation = xliff.xlifffile()
        translation.sourcelanguage = "en"
        translation.targetlanguage = "fr"

        for unit in source.units:
            translation.addunit(unit)

        # Verify structure is preserved
        serialized = bytes(translation)
        translation_reloaded = xliff.xlifffile.parsestring(serialized)
        assert len(translation_reloaded.units) == 2

        # First unit should be in group
        parent1 = translation_reloaded.units[0].xmlelement.getparent()
        assert etree.QName(parent1).localname == "group"

        # Second unit should be directly in body
        parent2 = translation_reloaded.units[1].xmlelement.getparent()
        assert etree.QName(parent2).localname == "body"

    def test_addunit_with_new_false(self) -> None:
        """Test that addunit with new=False doesn't duplicate units."""
        source_xliff = b"""<?xml version="1.0" encoding="utf-8"?>
<xliff version="1.1" xmlns="urn:oasis:names:tc:xliff:document:1.1">
    <file original="file1.txt" source-language="en" datatype="plaintext">
        <body>
            <group id="group1">
                <trans-unit id="hello">
                    <source>Hello</source>
                </trans-unit>
            </group>
        </body>
    </file>
</xliff>"""

        source = xliff.xlifffile.parsestring(source_xliff)
        translation = xliff.xlifffile()
        translation.sourcelanguage = "en"
        translation.targetlanguage = "fr"

        unit = source.units[0]

        # Add unit with new=False - should not add to XML tree
        translation.addunit(unit, new=False)

        # Unit should be in units list but namespace should be set
        assert len(translation.units) == 1
        assert unit.namespace == translation.namespace

        # But xmlelement should not be in the body or any group
        serialized = bytes(translation)
        # Should not contain the trans-unit since new=False
        translation_reloaded = xliff.xlifffile.parsestring(serialized)
        assert len(translation_reloaded.units) == 0

    def test_namespace_preservation_across_versions(self) -> None:
        """Test that namespace is properly handled when copying between XLIFF versions."""
        # XLIFF 1.1 source
        source_xliff = b"""<?xml version="1.0" encoding="utf-8"?>
<xliff version="1.1" xmlns="urn:oasis:names:tc:xliff:document:1.1">
    <file original="file1.txt" source-language="en" datatype="plaintext">
        <body>
            <group id="group1">
                <trans-unit id="hello">
                    <source>Hello</source>
                </trans-unit>
            </group>
        </body>
    </file>
</xliff>"""

        source = xliff.xlifffile.parsestring(source_xliff)
        translation = xliff.xlifffile()
        translation.sourcelanguage = "en"
        translation.targetlanguage = "fr"

        # Add unit and verify namespace is set correctly
        unit = source.units[0]
        translation.addunit(unit)

        # Namespace should be updated to target's namespace
        assert unit.namespace == translation.namespace

        # Verify it serializes correctly
        serialized = bytes(translation)
        assert b'xmlns="urn:oasis:names:tc:xliff:document:1.1"' in serialized

    def test_add_units_between_different_files(self) -> None:
        """Test adding units from one file with multiple <file> tags to another, preserving file structure."""
        # Source with two files
        source_xliff = b"""<?xml version="1.0" encoding="utf-8"?>
<xliff version="1.1" xmlns="urn:oasis:names:tc:xliff:document:1.1">
    <file original="source_file1.txt" source-language="en" datatype="plaintext">
        <body>
            <trans-unit id="hello">
                <source>Hello</source>
            </trans-unit>
        </body>
    </file>
    <file original="source_file2.txt" source-language="en" datatype="plaintext">
        <body>
            <trans-unit id="world">
                <source>World</source>
            </trans-unit>
        </body>
    </file>
</xliff>"""

        # Target with one file
        target_xliff = b"""<?xml version="1.0" encoding="utf-8"?>
<xliff version="1.1" xmlns="urn:oasis:names:tc:xliff:document:1.1">
    <file original="target_file.txt" source-language="en" datatype="plaintext" target-language="fr">
        <body>
            <trans-unit id="existing">
                <source>Existing</source>
                <target>Existent</target>
            </trans-unit>
        </body>
    </file>
</xliff>"""

        source = xliff.xlifffile.parsestring(source_xliff)
        target = xliff.xlifffile.parsestring(target_xliff)

        # Add units from source to target
        for unit in source.units:
            target.addunit(unit)

        # Verify all three files are present
        serialized = bytes(target)
        target_reloaded = xliff.xlifffile.parsestring(serialized)

        filenames = target_reloaded.getfilenames()
        assert len(filenames) == 3
        assert "target_file.txt" in filenames
        assert "source_file1.txt" in filenames
        assert "source_file2.txt" in filenames

        # Verify units are in correct files
        assert len(target_reloaded.units) == 3
        for unit in target_reloaded.units:
            unit_id = unit.getid()
            if "existing" in unit_id:
                assert unit_id.startswith("target_file.txt")
            elif "hello" in unit_id:
                assert unit_id.startswith("source_file1.txt")
            elif "world" in unit_id:
                assert unit_id.startswith("source_file2.txt")

    def test_group_preservation_across_different_namespaces(self) -> None:
        """Test that group structure is preserved when adding units, using namespace-independent matching."""
        # Source with groups
        source_xliff = b"""<?xml version="1.0" encoding="utf-8"?>
<xliff version="1.1" xmlns="urn:oasis:names:tc:xliff:document:1.1">
    <file original="source.txt" source-language="en" datatype="plaintext">
        <body>
            <group id="mygroup" restype="x-gettext-domain">
                <trans-unit id="hello">
                    <source>Hello</source>
                </trans-unit>
            </group>
        </body>
    </file>
</xliff>"""

        # Target without groups (but same namespace)
        target_xliff = b"""<?xml version="1.0" encoding="utf-8"?>
<xliff version="1.1" xmlns="urn:oasis:names:tc:xliff:document:1.1">
    <file original="target.txt" source-language="en" datatype="plaintext" target-language="fr">
        <body>
            <trans-unit id="existing">
                <source>Existing</source>
                <target>Existent</target>
            </trans-unit>
        </body>
    </file>
</xliff>"""

        source = xliff.xlifffile.parsestring(source_xliff)
        target = xliff.xlifffile.parsestring(target_xliff)

        # Verify initial state
        assert len(source.units) == 1
        assert len(target.units) == 1

        # Add unit with group from source to target
        target.addunit(source.units[0])

        # Verify target now has 2 units
        assert len(target.units) == 2

        # Serialize and reload to verify structure
        serialized = bytes(target)
        target_reloaded = xliff.xlifffile.parsestring(serialized)

        # Should still have 2 units after round-trip
        assert len(target_reloaded.units) == 2

        # Find the newly added unit by source text
        new_unit = None
        existing_unit = None
        for unit in target_reloaded.units:
            if unit.source == "Hello":
                new_unit = unit
            elif unit.source == "Existing":
                existing_unit = unit

        assert new_unit is not None, "New unit with 'Hello' not found"
        assert existing_unit is not None, "Existing unit not found"

        # Verify the new unit is in a group (namespace-independent check)
        new_parent = new_unit.xmlelement.getparent()
        assert etree.QName(new_parent).localname == "group", (
            "New unit should be in a group"
        )
        assert new_parent.get("id") == "mygroup"
        assert new_parent.get("restype") == "x-gettext-domain"

        # Verify the existing unit is NOT in a group
        existing_parent = existing_unit.xmlelement.getparent()
        assert etree.QName(existing_parent).localname == "body", (
            "Existing unit should be in body"
        )

        # Verify unit's namespace is correctly set to target's namespace
        assert new_unit.namespace == target.namespace

    def test_cross_namespace_group_and_file_preservation(self) -> None:
        """Test that groups and files are preserved when adding units across different XLIFF namespaces."""
        # Source with XLIFF 1.2 namespace, with groups
        source_xliff = b"""<?xml version="1.0" encoding="utf-8"?>
<xliff version="1.2" xmlns="urn:oasis:names:tc:xliff:document:1.2">
    <file original="source_file1.txt" source-language="en" datatype="plaintext">
        <body>
            <group id="group1" restype="x-gettext-domain">
                <trans-unit id="hello">
                    <source>Hello</source>
                </trans-unit>
            </group>
        </body>
    </file>
    <file original="source_file2.txt" source-language="en" datatype="plaintext">
        <body>
            <trans-unit id="world">
                <source>World</source>
            </trans-unit>
        </body>
    </file>
</xliff>"""

        # Target with XLIFF 1.1 namespace (different from source)
        target_xliff = b"""<?xml version="1.0" encoding="utf-8"?>
<xliff version="1.1" xmlns="urn:oasis:names:tc:xliff:document:1.1">
    <file original="target_file.txt" source-language="en" datatype="plaintext" target-language="fr">
        <body>
            <trans-unit id="existing">
                <source>Existing</source>
                <target>Existent</target>
            </trans-unit>
        </body>
    </file>
</xliff>"""

        source = xliff.xlifffile.parsestring(source_xliff)
        target = xliff.xlifffile.parsestring(target_xliff)

        # Verify different namespaces
        assert source.namespace == "urn:oasis:names:tc:xliff:document:1.2"
        assert target.namespace == "urn:oasis:names:tc:xliff:document:1.1"

        # Add units from source (1.2) to target (1.1)
        for unit in source.units:
            target.addunit(unit)

        # Check that units were registered
        assert len(target.units) == 3, f"Expected 3 units, got {len(target.units)}"

        # Verify files are preserved - target should have original file plus two from source
        filenames = target.getfilenames()
        assert len(filenames) == 3, (
            f"Expected 3 files, got {len(filenames)}: {filenames}"
        )
        assert "target_file.txt" in filenames
        assert "source_file1.txt" in filenames
        assert "source_file2.txt" in filenames

        # Serialize and check structure
        serialized = bytes(target)

        # Parse with lxml to check namespace and structure

        tree = etree.fromstring(serialized)

        # Verify the document uses target's namespace (1.1)
        assert tree.nsmap[None] == "urn:oasis:names:tc:xliff:document:1.1"

        # Check for groups in target namespace
        groups = tree.xpath(
            "//ns:group", namespaces={"ns": "urn:oasis:names:tc:xliff:document:1.1"}
        )
        assert len(groups) >= 1, f"Expected at least 1 group, found {len(groups)}"

        # Verify group attributes are preserved
        group1 = None
        for g in groups:
            if g.get("id") == "group1":
                group1 = g
                break
        assert group1 is not None, "group1 not found"
        assert group1.get("restype") == "x-gettext-domain"

        # Verify all files are present in target namespace
        files = tree.xpath(
            "//ns:file", namespaces={"ns": "urn:oasis:names:tc:xliff:document:1.1"}
        )
        assert len(files) == 3, f"Expected 3 file elements, found {len(files)}"

        file_originals = [f.get("original") for f in files]
        assert "target_file.txt" in file_originals
        assert "source_file1.txt" in file_originals
        assert "source_file2.txt" in file_originals

    def test_indent(self) -> None:
        xlfsource = b"""<?xml version="1.0" encoding="UTF-8"?>
<xliff xmlns="urn:oasis:names:tc:xliff:document:1.1" version="1.1">
  <file original="doc.txt" source-language="en-US">
    <body>
      <trans-unit id="1" xml:space="preserve">
        <source>File</source>
        <target/>
      </trans-unit>
    </body>
  </file>
</xliff>
"""
        xlfsourcenote = b"""<?xml version="1.0" encoding="UTF-8"?>
<xliff xmlns="urn:oasis:names:tc:xliff:document:1.1" version="1.1">
  <file original="doc.txt" source-language="en-US">
    <body>
      <trans-unit id="1" xml:space="preserve">
        <source>File</source>
        <target/>
        <note>Test note</note>
      </trans-unit>
    </body>
  </file>
</xliff>
"""
        xfile = xliff.xlifffile.parsestring(xlfsource)
        assert bytes(xfile) == xlfsource
        xfile.units[0].addnote("Test note")
        assert bytes(xfile) == xlfsourcenote

    def test_add_target(self) -> None:
        xlfsource = """<?xml version="1.0" encoding="UTF-8"?>
<xliff xmlns="urn:oasis:names:tc:xliff:document:1.1" version="1.1">
  <file original="doc.txt" source-language="en-US">
    <body>
      <trans-unit id="1" xml:space="preserve">
      </trans-unit>
    </body>
  </file>
</xliff>
"""
        xlftarget = """<?xml version="1.0" encoding="UTF-8"?>
<xliff xmlns="urn:oasis:names:tc:xliff:document:1.1" version="1.1">
  <file original="doc.txt" source-language="en-US">
    <body>
      <trans-unit id="1" xml:space="preserve">
        <target>Soubor</target>
      </trans-unit>
    </body>
  </file>
</xliff>
"""
        xfile = xliff.xlifffile.parsestring(xlfsource)
        assert bytes(xfile).decode() == xlfsource
        xfile.units[0].rich_target = ["Soubor"]
        assert bytes(xfile).decode("ascii") == xlftarget

    def test_preserve(self) -> None:
        xlfsource = """<?xml version="1.0" encoding="UTF-8"?>
<xliff xmlns="urn:oasis:names:tc:xliff:document:1.1" version="1.1">
  <file original="doc.txt" source-language="en-US">
    <body>
      <trans-unit id="1" translate="yes">
        <source>Hello</source>
        <target/>
      </trans-unit>
    </body>
  </file>
</xliff>
"""
        xfile = xliff.xlifffile.parsestring(xlfsource)
        assert bytes(xfile).decode() == xlfsource
        xfile.units[0].target = "H  E"
        newfile = xliff.xlifffile.parsestring(bytes(xfile))
        assert newfile.units[0].target == "H  E"

    def test_closing_tags(self) -> None:
        xlfsource = """<?xml version="1.0" encoding="utf-8"?>
<xliff xmlns="urn:oasis:names:tc:xliff:document:1.2" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" version="1.2" xsi:schemaLocation="urn:oasis:names:tc:xliff:document:1.2 xliff-core-1.2-transitional.xsd">
  <file datatype="xml" source-language="en-US" target-language="en-US" original="Email - SMTP API">
    <body>
      <group id="body">
        <trans-unit id="Codeunit 270637162 - NamedType 3430817766" maxwidth="0" size-unit="char" translate="yes" xml:space="preserve">
          <source>Please connect to a server first.</source>
          <target state="translated">Please connect to a server first.</target>
          <note from="Developer" annotates="general" priority="2"></note>
          <note from="Xliff Generator" annotates="general" priority="3">Codeunit MailKit Client - NamedType ServerNotConnectedErr</note>
        </trans-unit>
      </group>
    </body>
  </file>
</xliff>
"""
        xlifffile = xliff.xlifffile.parsestring(xlfsource)
        xlifffile.XMLSelfClosingTags = False
        xlifffile.XMLuppercaseEncoding = False

        assert bytes(xlifffile).decode("utf-8") == xlfsource

    def test_context_groups(self) -> None:
        xlfsource = """<?xml version="1.0" encoding="utf-8"?>
<xliff xmlns="urn:oasis:names:tc:xliff:document:1.2" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" version="1.2" xsi:schemaLocation="urn:oasis:names:tc:xliff:document:1.2 xliff-core-1.2-transitional.xsd">
  <file datatype="xml" source-language="en-US" target-language="en-US" original="Email - SMTP API">
    <body>
      <group id="body">
        <trans-unit id="Codeunit 270637162 - NamedType 3430817766" maxwidth="0" size-unit="char" translate="yes" xml:space="preserve">
          <source>Please connect to a server first.</source>
          <target state="translated">Please connect to a server first.</target>
          <context-group purpose="location">
            <context context-type="sourcefile">some-directory/on-some-test/test.file</context>
            <context context-type="linenumber">222</context>
          </context-group>
        </trans-unit>
      </group>
    </body>
  </file>
</xliff>
"""
        xlifffile = xliff.xlifffile.parsestring(xlfsource)
        contextGroups = xlifffile.units[0].getcontextgroupsbyattribute(
            "purpose", "location"
        )

        assert contextGroups[0][0][0] == "sourcefile"
        assert contextGroups[0][0][1] == "some-directory/on-some-test/test.file"

        assert contextGroups[0][1][0] == "linenumber"
        assert contextGroups[0][1][1] == "222"

    def test_getlocations(self) -> None:
        xlfsource = """<?xml version="1.0" encoding="utf-8"?>
<xliff xmlns="urn:oasis:names:tc:xliff:document:1.2" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" version="1.2" xsi:schemaLocation="urn:oasis:names:tc:xliff:document:1.2 xliff-core-1.2-transitional.xsd">
  <file datatype="xml" source-language="en-US" target-language="en-US" original="Email - SMTP API">
    <body>
      <group id="body">
        <trans-unit id="Codeunit 270637162 - NamedType 3430817766" maxwidth="0" size-unit="char" translate="yes" xml:space="preserve">
          <source>Please connect to a server first.</source>
          <target state="translated">Please connect to a server first.</target>
          <context-group purpose="location">
            <context context-type="sourcefile">some-directory/on-some-test/test.file</context>
            <context context-type="linenumber">222</context>
          </context-group>
        </trans-unit>
      </group>
    </body>
  </file>
</xliff>
"""
        xlifffile = xliff.xlifffile.parsestring(xlfsource)
        locations = xlifffile.units[0].getlocations()

        assert locations == ["some-directory/on-some-test/test.file:222"]

    def test_addlocation(self) -> None:
        xlfsource = """<?xml version="1.0" encoding="utf-8"?>
<xliff xmlns="urn:oasis:names:tc:xliff:document:1.2" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" version="1.2" xsi:schemaLocation="urn:oasis:names:tc:xliff:document:1.2 xliff-core-1.2-transitional.xsd">
  <file datatype="xml" source-language="en-US" target-language="en-US" original="Email - SMTP API">
    <body>
      <group id="body">
        <trans-unit id="Codeunit 270637162 - NamedType 3430817766" maxwidth="0" size-unit="char" translate="yes" xml:space="preserve">
          <source>Please connect to a server first.</source>
          <target state="translated">Please connect to a server first.</target>
          <context-group purpose="location">
            <context context-type="sourcefile">some-directory/on-some-test/test.file</context>
            <context context-type="linenumber">222</context>
          </context-group>
        </trans-unit>
      </group>
    </body>
  </file>
</xliff>
"""
        xlifffile = xliff.xlifffile.parsestring(xlfsource)
        xlifffile.units[0].addlocation("some-directory/on-some-test/test.file2:333")
        xlifffile.units[0].addlocation("docs/config.md:block 1 (header)")
        xlifffile.units[0].addlocation("file.rst")
        locations = xlifffile.units[0].getlocations()
        assert locations == [
            "some-directory/on-some-test/test.file:222",
            "some-directory/on-some-test/test.file2:333",
            "docs/config.md:block 1 (header)",
            "file.rst",
        ]

    def test_huge(self) -> None:
        # This create approx 120MB Xliff file
        xlfsource = self.skeleton % (
            """<trans-unit id="1">
                     <source>text</source>
                     <target/>
                 </trans-unit>"""
            * 100_000
        )
        xlifffile = xliff.xlifffile.parsestring(xlfsource)
        assert len(xlifffile.units) == 100_000

    def test_preserve_add(self) -> None:
        # spellchecker:off
        target = """Por favor, ten a la mano los siguientes documentos para completar el proceso:
• INE (Credencial para votar) o pasaporte, vigentes
• Carátula del estado de cuenta bancario personal
• Cuenta CLABE Bancaria
• Constancia de Situación Fiscal con el Régimen de Sueldos y Salarios e Ingresos Asimilados a Salarios, con una vigencia no mayor a 90 días, completa, en FORMATO PDF"""
        # spellchecker:on
        xlfsource_template = """<?xml version="1.0" encoding="UTF-8"?>
<xliff xmlns="urn:oasis:names:tc:xliff:document:1.2" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" version="1.2" xsi:schemaLocation="urn:oasis:names:tc:xliff:document:1.2 xliff-core-1.2-transitional.xsd">
  <file datatype="xml" source-language="en-US" target-language="en-US" original="Email - SMTP API">
    <body>
      <group id="body">
        <trans-unit id="8864d0bd25f99594bbfb59780cb19a091946d8e6" datatype="html"{extra}>
          <source>Status</source>{target}
          <context-group purpose="location">
            <context context-type="sourcefile">src/app/settings/component/profile/example/example.component.html</context>
            <context context-type="linenumber">3</context>
          </context-group>
          <note priority="1" from="description">settings</note>
        </trans-unit>
      </group>
    </body>
  </file>
</xliff>
"""
        target_element = f"""
          <target state="translated">{target}</target>"""
        xlfsource_blank = xlfsource_template.format(target="", extra="")
        xlfsource_plain = xlfsource_template.format(target=target_element, extra="")
        xlfsource_preserve = xlfsource_template.format(
            target=target_element, extra=' xml:space="preserve"'
        )
        xlifffile = xliff.xlifffile.parsestring(xlfsource_plain)

        assert len(xlifffile.units) == 1
        unit = xlifffile.units[0]

        assert unit.target == target.replace("\n", " ")
        unit.target = target

        assert bytes(xlifffile).decode() == xlfsource_preserve

        xlifffile = xliff.xlifffile.parsestring(xlfsource_blank)

        assert len(xlifffile.units) == 1
        unit = xlifffile.units[0]

        assert unit.target is None
        unit.target = target

        assert bytes(xlifffile).decode() == xlfsource_template.format(
            target=target_element, extra=' xml:space="preserve" approved="yes"'
        )

        xlifffile = xliff.xlifffile.parsestring(xlfsource_preserve)

        assert len(xlifffile.units) == 1
        unit = xlifffile.units[0]

        assert unit.target == target

        assert bytes(xlifffile).decode() == xlfsource_preserve

        xlifffile = xliff.xlifffile.parsestring(xlfsource_plain)

        assert len(xlifffile.units) == 1
        unit = xlifffile.units[0]

        assert unit.target == target.replace("\n", " ")
        unit.rich_target = [target]

        assert bytes(xlifffile).decode() == xlfsource_preserve

    def test_xliff1_rejects_xliff2_files(self) -> None:
        """Test that XLIFF 1.x parser rejects XLIFF 2.0 files with a clear error."""
        xliff2_content = b"""<?xml version="1.0" encoding="utf-8"?>
<xliff version="2.0" xmlns="urn:oasis:names:tc:xliff:document:2.0" srcLang="en">
    <file id="f1">
        <unit id="1">
            <segment>
                <source>Hello</source>
                <target>Hola</target>
            </segment>
        </unit>
    </file>
</xliff>"""
        with pytest.raises(ValueError) as exc_info:
            xliff.xlifffile.parsestring(xliff2_content)
        assert "XLIFF 2" in str(exc_info.value)
        assert "version='2.0'" in str(exc_info.value)
        assert "xliff2.Xliff2File" in str(exc_info.value)

    def test_xliff1_rejects_xliff2_files_without_version(self) -> None:
        """Test that XLIFF 1.x parser rejects XLIFF 2.0 files without version attribute."""
        xliff2_no_version = b"""<?xml version="1.0" encoding="utf-8"?>
<xliff xmlns="urn:oasis:names:tc:xliff:document:2.0" srcLang="en">
    <file id="f1">
        <unit id="1">
            <segment>
                <source>Hello</source>
                <target>Hola</target>
            </segment>
        </unit>
    </file>
</xliff>"""
        with pytest.raises(ValueError) as exc_info:
            xliff.xlifffile.parsestring(xliff2_no_version)
        assert "XLIFF 2" in str(exc_info.value)
        assert "namespace" in str(exc_info.value)
        assert "xliff2.Xliff2File" in str(exc_info.value)



from io import BytesIO

from pytest import mark, raises

from translate.misc.multistring import multistring
from translate.storage import base, jsonl10n

from . import test_monolingual

JSON_I18NEXT = """{
    "key": "value",
    "keyDeep": {
        "inner": "value"
    },
    "keyPluralSimple": "the singular",
    "keyPluralSimple_plural": "the plural",
    "keyPluralMultipleEgArabic_0": "the plural form 0",
    "keyPluralMultipleEgArabic_1": "the plural form 1",
    "keyPluralMultipleEgArabic_2": "the plural form 2",
    "keyPluralMultipleEgArabic_3": "the plural form 3",
    "keyPluralMultipleEgArabic_4": "the plural form 4",
    "keyPluralMultipleEgArabic_5": "the plural form 5"
}
"""

JSON_I18NEXT_V4 = """{
    "key": "value",
    "keyDeep": {
        "inner": "value"
    },
    "keyPluralSimple_one": "the singular",
    "keyPluralSimple_other": "the plural",
    "keyPluralMultipleEgArabic_zero": "the plural form 0",
    "keyPluralMultipleEgArabic_one": "the plural form 1",
    "keyPluralMultipleEgArabic_two": "the plural form 2",
    "keyPluralMultipleEgArabic_few": "the plural form 3",
    "keyPluralMultipleEgArabic_many": "the plural form 4",
    "keyPluralMultipleEgArabic_other": "the plural form 5"
}
"""

JSON_I18NEXT_V4_FIXED = """{
    "key": "value",
    "keyDeep": {
        "inner": "value"
    },
    "keyPluralSimple_zero": "",
    "keyPluralSimple_one": "the singular",
    "keyPluralSimple_two": "",
    "keyPluralSimple_few": "",
    "keyPluralSimple_many": "",
    "keyPluralSimple_other": "the plural",
    "keyPluralMultipleEgArabic_zero": "the plural form 0",
    "keyPluralMultipleEgArabic_one": "the plural form 1",
    "keyPluralMultipleEgArabic_two": "the plural form 2",
    "keyPluralMultipleEgArabic_few": "the plural form 3",
    "keyPluralMultipleEgArabic_many": "the plural form 4",
    "keyPluralMultipleEgArabic_other": "the plural form 5"
}
"""

JSON_FLAT_I18NEXT_V4 = """{
    "key": "value",
    "keyDeep.inner": "value",
    "keyPluralSimple_one": "the singular",
    "keyPluralSimple_other": "the plural",
    "keyPluralMultipleEgArabic_zero": "the plural form 0",
    "keyPluralMultipleEgArabic_one": "the plural form 1",
    "keyPluralMultipleEgArabic_two": "the plural form 2",
    "keyPluralMultipleEgArabic_few": "the plural form 3",
    "keyPluralMultipleEgArabic_many": "the plural form 4",
    "keyPluralMultipleEgArabic_other": "the plural form 5"
}
"""

JSON_FLAT_I18NEXT_V4_FIXED = """{
    "key": "value",
    "keyDeep.inner": "value",
    "keyPluralSimple_zero": "",
    "keyPluralSimple_one": "the singular",
    "keyPluralSimple_two": "",
    "keyPluralSimple_few": "",
    "keyPluralSimple_many": "",
    "keyPluralSimple_other": "the plural",
    "keyPluralMultipleEgArabic_zero": "the plural form 0",
    "keyPluralMultipleEgArabic_one": "the plural form 1",
    "keyPluralMultipleEgArabic_two": "the plural form 2",
    "keyPluralMultipleEgArabic_few": "the plural form 3",
    "keyPluralMultipleEgArabic_many": "the plural form 4",
    "keyPluralMultipleEgArabic_other": "the plural form 5"
}
"""

JSON_FLAT_I18NEXT_V4_PLURAL = b"""{
    "key": "value",
    "keyDeep.inner": "value",
    "keyPluralSimple": "Ahoj",
    "keyPluralMultipleEgArabic": "Nazdar"
}
"""

JSON_I18NEXT_PLURAL = b"""{
    "key": "value",
    "keyDeep": {
        "inner": "value"
    },
    "keyPluralSimple": "Ahoj",
    "keyPluralMultipleEgArabic": "Nazdar"
}
"""

JSON_I18NEXT_NESTED_ARRAY_BEFORE = """{
    "apps": {
        "legacy": "app1"
    }
}
"""
JSON_I18NEXT_NESTED_ARRAY = """{
    "apps": [
        {
            "title": "app1",
            "description": "test description"
        },
        {
            "title": "app2",
            "description": "test description 2"
        }
    ]
}
"""
JSON_ARRAY = b"""{
    "key": [
        "One",
        "Two",
        "Three"
    ]
}
"""
JSON_COMPLEX = b"""{
    "key": "value",
    "key.key": "value",
    "key[0]": "value2",
    "key3": [
        "one",
        "two"
    ],
    "key4": [
        {
            "nested": "one"
        },
        [
            "one",
            "two"
        ]
    ]
}
"""
# spellchecker:off
JSON_COMPLEX_ARRAY = r"""[
    {
        "url": "massivholztische",
        "heading": "Welche Tischgröße ist für mich die richtige?",
        "content": "Überlege zunächst wie viele Personen regelmäßig\r\n"
    },
    {
        "url": "massivholztische",
        "heading": "Bietet Öl genügend Schutz für meinen Massivholztisch?",
        "content": "<p>Holzöl bietet bei normaler Nutzung"
    }
]
""".encode()
# spellchecker:on
JSON_GOTEXT = b"""{
    "language": "en-US",
    "messages": [
        {
            "id": "tag",
            "message": "{N} tags",
            "translatorComment": "a piece or strip of strong paper, plastic, metal, leather, etc., for attaching by one end to something as a mark or label",
            "translation": {
                "select": {
                    "feature": "plural",
                    "arg": "N",
                    "cases": {
                        "one": {
                            "msg": "{N} tag"
                        },
                        "other": {
                            "msg": "{N} tags"
                        }
                    }
                }
            },
            "placeholders": [
                {
                    "id": "N",
                    "string": "%[1]d",
                    "type": "int",
                    "underlyingType": "int",
                    "argNum": 1,
                    "expr": "n"
                }
            ]
        },
        {
            "id": "table",
            "message": "Table",
            "translatorComment": "an article of furniture consisting of a flat, slablike top supported on one or more legs or other supports",
            "translation": "Table"
        }
    ]
}
"""
JSON_GOI18N = b"""[
    {
        "id": "tag",
        "description": "a piece or strip of strong paper, plastic, metal, leather, etc., for attaching by one end to something as a mark or label",
        "translation": {
            "one": "{{.count}} tag",
            "other": "{{.count}} tags"
        }
    },
    {
        "id": "table",
        "description": "an article of furniture consisting of a flat, slablike top supported on one or more legs or other supports",
        "translation": "Table"
    }
]
"""

JSON_GOI18N_V2 = """{
    "simple": "value",
    "plural": {
        "zero": "the plural form 0",
        "one": "the plural form 1",
        "two": "the plural form 2",
        "few": "the plural form 3",
        "many": "the plural form 4",
        "other": "the plural form 5"
    }
}
"""

JSON_GOI18N_V2_PLURAL = """{
    "simple": "value",
    "plural": "plural"
}
"""


JSON_GOI18N_V2_SIMPLE = """{
    "key": "value",
    "table": {
        "description": "an article of furniture consisting of a flat, slablike top supported on one or more legs or other supports",
        "other": "Table"
    },
    "tag": {
        "description": "a piece or strip of strong paper, plastic, metal, leather, etc., for attaching by one end to something as a mark or label",
        "one": "{{.count}} tag",
        "other": "{{.count}} tags"
    },
    "nested.key.hello": "world",
    "nested.key.world": "hello",
    ".world": "hello"
}
"""

JSON_GOI18N_V2_COMPLEX = """{
    "key": {
        "other": "value"
    },
    "table": {
        "description": "an article of furniture consisting of a flat, slablike top supported on one or more legs or other supports",
        "other": "Table"
    },
    "tag": {
        "description": "a piece or strip of strong paper, plastic, metal, leather, etc., for attaching by one end to something as a mark or label",
        "one": "{{.count}} tag",
        "other": "{{.count}} tags"
    },
    "nested.key.hello": {
        "other": "world"
    },
    "nested.key.world": {
        "other": "hello"
    },
    ".world": {
        "other": "hello"
    }
}
"""

JSON_ARB = b"""{
  "@@last_modified": "2019-11-06T22:41:37.002648",
  "Back": "Back",
  "@Back": {
    "type": "text",
    "placeholders": {}
  },
  "Next": "Next",
  "@Next": {
    "type": "text",
    "placeholders": {}
  },
  "Done": "Done",
  "@Done": {
    "type": "text",
    "placeholders": {}
  }
}
"""


class TestJSONResourceUnit(test_monolingual.TestMonolingualUnit):
    UnitClass = jsonl10n.BaseJsonUnit


class TestJSONResourceStore(test_monolingual.TestMonolingualStore):
    StoreClass = jsonl10n.JsonFile

    def test_serialize(self) -> None:
        store = self.StoreClass()
        store.parse('{"key": "value"}')
        out = BytesIO()
        store.serialize(out)

        assert out.getvalue() == b'{\n    "key": "value"\n}\n'

    def test_can_not_detect(self) -> None:
        store = self.StoreClass()
        with raises(base.ParseError):
            store.parse(
                b"PK\x03\x04\x14\x00\x06\x00\x08\x00\x00\x00!\x00b\xee\x9dh^\x01\x00\x00\x90\x04\x00\x00\x13\x00\x08\x02"
            )

    def test_error(self) -> None:
        store = self.StoreClass()
        with raises(base.ParseError):
            store.parse('{"key": "value"')

    def test_filter(self) -> None:
        store = self.StoreClass(filter=["key"])
        store.parse('{"key": "value", "other": "second"}')
        assert len(store.units) == 1
        assert store.units[0].source == "value"

    def test_ordering(self) -> None:
        store = self.StoreClass()
        store.parse(
            """{
    "foo": "foo",
    "bar": "bar",
    "baz": "baz"
}"""
        )

        assert store.units[0].source == "foo"
        assert store.units[2].source == "baz"

    def test_args(self) -> None:
        store = self.StoreClass()
        store.parse(
            """{
    "foo": "foo",
    "bar": "bar",
    "baz": "baz"
}"""
        )
        store.dump_args["sort_keys"] = True

        out = BytesIO()
        store.serialize(out)

        assert (
            out.getvalue()
            == b"""{
    "bar": "bar",
    "baz": "baz",
    "foo": "foo"
}
"""
        )

    def test_bom(self) -> None:
        content = "{}\n".encode("utf-8-sig")
        store = self.StoreClass()
        store.parse(content)
        assert len(store.units) == 0
        out = BytesIO()
        store.serialize(out)
        assert out.getvalue() == content

    def test_complex(self) -> None:
        store = self.StoreClass()
        store.parse(JSON_COMPLEX)

        out = BytesIO()
        store.serialize(out)

        assert out.getvalue() == JSON_COMPLEX

    def test_complex_array(self) -> None:
        store = self.StoreClass()
        store.parse(JSON_COMPLEX_ARRAY)

        out = BytesIO()
        store.serialize(out)

        assert out.getvalue() == JSON_COMPLEX_ARRAY

    def test_add(self) -> None:
        expected = """{
    "simple.key": "source"
}
"""
        store = self.StoreClass()

        unit = self.StoreClass.UnitClass(
            "source",
        )
        assert str(unit) != expected.strip()
        unit.setid("simple.key")
        store.addunit(unit)

        assert str(unit) == expected.strip()
        assert bytes(store).decode() == expected

    def test_add_list_like(self) -> None:
        store = self.StoreClass()

        unit = self.StoreClass.UnitClass("source")
        unit.setid("[0]")
        store.addunit(unit)

        unit = self.StoreClass.UnitClass("source2")
        unit.setid("key")
        store.addunit(unit)

        assert (
            bytes(store).decode()
            == """{
    "[0]": "source",
    "key": "source2"
}
"""
        )
        unit.setid("[1]")
        assert (
            bytes(store).decode()
            == """{
    "[0]": "source",
    "[1]": "source2"
}
"""
        )

    def test_add_blank(self) -> None:
        expected = """{
    "simple.key": "source"
}
"""
        store = self.StoreClass()
        store.parse("true")

        unit = self.StoreClass.UnitClass(
            "source",
        )
        assert str(unit) != expected.strip()
        unit.setid("simple.key")
        store.addunit(unit)

        assert str(unit) == expected.strip()
        assert bytes(store).decode() == expected

    def test_types(self) -> None:
        store = self.StoreClass()
        store.parse('{"key": 1}')
        assert store.units[0].target == 1
        store.units[0].target = 2

        assert (
            bytes(store).decode()
            == """{
    "key": 2
}
"""
        )
        store.units[0].target = "two"

        assert (
            bytes(store).decode()
            == """{
    "key": "two"
}
"""
        )

    def test_null(self) -> None:
        jsondata = """{
    "foo": [
        null,
        null,
        "Text"
    ]
}
"""
        store = self.StoreClass()
        store.parse(jsondata)
        assert len(store.units) == 3

        assert store.units[0].target is None
        assert store.units[1].target is None
        assert store.units[2].target == "Text"

        out = BytesIO()
        store.serialize(out)

        assert out.getvalue().decode() == jsondata


class TestJSONNestedResourceStore(test_monolingual.TestMonolingualUnit):
    StoreClass = jsonl10n.JsonNestedFile

    def test_serialize(self) -> None:
        store = self.StoreClass()
        store.parse('{"key": {"second": "value"}}')
        out = BytesIO()
        store.serialize(out)

        assert (
            out.getvalue() == b'{\n    "key": {\n        "second": "value"\n    }\n}\n'
        )

    def test_ordering(self) -> None:
        # spellchecker:off
        data = b"""{
    "foo": "foo",
    "bar": {
        "ba1": "bag",
        "ba2": "bag",
        "ba3": "bag",
        "ba4": "baz"
    }
}
"""
        store = self.StoreClass()
        store.parse(data)

        assert store.units[0].source == "foo"
        assert store.units[1].getid() == ".bar.ba1"
        assert store.units[2].getid() == ".bar.ba2"
        assert store.units[3].getid() == ".bar.ba3"
        assert store.units[4].getid() == ".bar.ba4"
        # spellchecker:on

        out = BytesIO()
        store.serialize(out)

        assert out.getvalue() == data

    def test_array(self) -> None:
        store = self.StoreClass()
        store.parse(JSON_ARRAY)

        assert store.units[0].source == "One"
        assert store.units[0].getid() == ".key[0]"
        assert store.units[1].getid() == ".key[1]"

        out = BytesIO()
        store.serialize(out)

        assert out.getvalue() == JSON_ARRAY

    def test_add(self) -> None:
        store = self.StoreClass()

        unit = self.StoreClass.UnitClass("source")
        unit.setid("simple.key")
        store.addunit(unit)

        expected = """{
    "simple": {
        "key": "source"
    }
}
"""
        assert bytes(store).decode() == expected

    def test_add_index(self) -> None:
        store = self.StoreClass()

        unit = self.StoreClass.UnitClass("source")
        unit.setid("simple.list[2].key")
        store.addunit(unit)

        expected = """{
    "simple": {
        "list": [
            {},
            {},
            {
                "key": "source"
            }
        ]
    }
}
"""
        assert bytes(store).decode() == expected

    def test_add_index_nested(self) -> None:
        store = self.StoreClass()
        store.parse('{"foo":[["x", "y"]]}')

        assert len(store.units) == 2
        assert store.units[0].getid() == ".foo[0][0]"
        assert store.units[1].getid() == ".foo[0][1]"

        unit = self.StoreClass.UnitClass("source")
        unit.setid("values[2][0]")
        store.addunit(unit)

        expected = """{
    "foo": [
        [
            "x",
            "y"
        ]
    ],
    "values": [
        [],
        [],
        [
            "source"
        ]
    ]
}
"""
        assert bytes(store).decode() == expected

    def test_nested_list_mixed(self) -> None:
        data = """{
    "story_9795": {
        "tsr_0": [
            [
                "‥",
                "Combination Carcer Tullianum & parco"
            ],
            "Archeologico del Colosseo"
        ]
    }
}
"""
        store = self.StoreClass()
        store.parse(data)
        assert len(store.units) == 3
        assert store.units[0].getid() == ".story_9795.tsr_0[0][0]"
        assert store.units[1].getid() == ".story_9795.tsr_0[0][1]"
        assert store.units[2].getid() == ".story_9795.tsr_0[1]"

        assert bytes(store).decode() == data

        store = self.StoreClass()
        unit = self.StoreClass.UnitClass("Archeologico del Colosseo")
        unit.setid(".story_9795.tsr_0[1]")
        store.addunit(unit)
        unit = self.StoreClass.UnitClass("Combination Carcer Tullianum & parco")
        unit.setid(".story_9795.tsr_0[0][1]")
        store.addunit(unit)
        unit = self.StoreClass.UnitClass("‥")
        unit.setid(".story_9795.tsr_0[0][0]")
        store.addunit(unit)

        assert bytes(store).decode() == data

    def test_list_to_dict(self) -> None:
        data = """{
    "userInfoPage": [
        "Name"
    ]
}
"""
        store = self.StoreClass()
        store.parse(data)
        assert len(store.units) == 1
        assert bytes(store).decode() == data

        unit = self.StoreClass.UnitClass("Test")
        unit.setid("userInfoPage.nesting")
        store.addunit(unit)
        assert (
            bytes(store).decode()
            == """{
    "userInfoPage": {
        "0": "Name",
        "nesting": "Test"
    }
}
"""
        )

    def test_complex_keys(self) -> None:
        data = """{
    "view": {
        "([^.,0-9]|^)1([^.,0-9]|$)": "View `x` comment",
        "": "View `x` comments"
    }
}
"""
        store = self.StoreClass()
        store.parse(data)
        assert len(store.units) == 2
        assert bytes(store).decode() == data

    def test_add_other(self) -> None:
        jsontext = """{
    "simple.key": "source"
}
"""
        store = self.StoreClass()
        store.parse(jsontext)

        assert len(store.units) == 1
        assert store.units[0].source == "source"
        assert store.units[0].getid() == ".simple.key"

        assert bytes(store).decode() == jsontext

        newstore = self.StoreClass()
        newstore.addunit(store.units[0])

        assert bytes(newstore).decode() == jsontext

    @mark.parametrize(
        ("id_string", "expected"),
        [
            ("[0]", [("index", 0)]),
            ("test[0]", [("key", "test"), ("index", 0)]),
            (
                "test[0][1][2][3]",
                [
                    ("key", "test"),
                    ("index", 0),
                    ("index", 1),
                    ("index", 2),
                    ("index", 3),
                ],
            ),
            ("[test]selection", [("key", "[test]selection")]),
            ("[test][0]selection", [("key", "[test][0]selection")]),
            ("[0][test]selection", [("index", 0), ("key", "[test]selection")]),
            ("", [("key", "")]),
        ],
    )
    def test_from_string(self, id_string, expected) -> None:
        id_class = self.StoreClass.UnitClass.IdClass
        from_string = id_class.from_string
        assert from_string(id_string) == id_class(expected)

    def test_dot_keys(self) -> None:
        jsontext = """{
    "key.dot": "message"
}
"""
        store = self.StoreClass()
        store.parse(jsontext)
        # Edit target
        store.units[0].target = "message"
        assert bytes(store).decode() == jsontext


class TestWebExtensionUnit(test_monolingual.TestMonolingualUnit):
    UnitClass = jsonl10n.WebExtensionJsonUnit


class TestWebExtensionStore(test_monolingual.TestMonolingualStore):
    StoreClass = jsonl10n.WebExtensionJsonFile

    def test_serialize(self) -> None:
        store = self.StoreClass()
        store.parse('{"key": {"message": "value", "description": "note"}}')
        out = BytesIO()
        store.serialize(out)

        assert (
            out.getvalue()
            == b'{\n    "key": {\n        "message": "value",\n        "description": "note"\n    }\n}\n'
        )

    def test_dot_keys(self) -> None:
        store = self.StoreClass()
        store.parse('{"key.dot": {"message": "value", "description": "note"}}')
        out = BytesIO()
        store.serialize(out)

        assert (
            out.getvalue()
            == b'{\n    "key.dot": {\n        "message": "value",\n        "description": "note"\n    }\n}\n'
        )

    def test_leading_dot_keys(self) -> None:
        jsontext = """{
    ".dot": {
        "message": "value",
        "description": "note"
    }
}
"""
        store = self.StoreClass()
        store.parse(jsontext)
        assert bytes(store).decode() == jsontext

    def test_serialize_no_description(self) -> None:
        store = self.StoreClass()
        store.parse('{"key": {"message": "value"}}')
        out = BytesIO()
        store.serialize(out)

        assert (
            out.getvalue() == b'{\n    "key": {\n        "message": "value"\n    }\n}\n'
        )

    def test_set_target(self) -> None:
        store = self.StoreClass()
        store.parse('{"key": {"message": "value", "description": "note"}}')
        store.units[0].target = "another"
        out = BytesIO()
        store.serialize(out)

        assert (
            out.getvalue()
            == b'{\n    "key": {\n        "message": "another",\n        "description": "note"\n    }\n}\n'
        )

    def test_placeholders(self) -> None:
        DATA = """{
    "youCanClose": {
        "message": "Bravo ! Votre compte $SITE$ est relié à Scrobbly. Vous pouvez fermer et revenir en arrière",
        "placeholders": {
            "site": {
                "content": "$1",
                "example": "AniList"
            }
        }
    }
}
""".encode()

        store = self.StoreClass()
        store.parse(DATA)
        assert store.units[0].placeholders is not None
        out = BytesIO()
        store.serialize(out)

        assert out.getvalue() == DATA

    def test_comments(self) -> None:
        DATA = """{
    "test": {
        // Comment
        "message": "Message //"
    },
    "test//": { // Comment
        "message": "Message // Other"
    }
}
"""
        store = self.StoreClass()
        store.parse(DATA)
        assert len(store.units) == 2
        # Any comments will be stripped
        assert (
            bytes(store).decode()
            == """{
    "test": {
        "message": "Message //"
    },
    "test//": {
        "message": "Message // Other"
    }
}
"""
        )


class TestI18NextStore(test_monolingual.TestMonolingualStore):
    StoreClass = jsonl10n.I18NextFile

    def test_serialize(self) -> None:
        store = self.StoreClass()
        store.settargetlanguage("ar")
        store.parse(JSON_I18NEXT)
        out = BytesIO()
        store.serialize(out)

        assert out.getvalue().decode() == JSON_I18NEXT

    def test_units(self) -> None:
        store = self.StoreClass()
        store.settargetlanguage("ar")
        store.parse(JSON_I18NEXT)
        assert len(store.units) == 4

    def test_plurals(self) -> None:
        store = self.StoreClass()
        store.settargetlanguage("ar")
        store.parse(JSON_I18NEXT)

        # Remove plurals
        store.units[2].target = "Ahoj"
        store.units[3].target = "Nazdar"
        out = BytesIO()
        store.serialize(out)

        assert out.getvalue() == JSON_I18NEXT_PLURAL

        # Bring back plurals
        store.units[2].target = multistring(
            [
                "the singular",
                "the plural",
            ]
        )
        store.units[3].target = multistring(
            [
                "the plural form 0",
                "the plural form 1",
                "the plural form 2",
                "the plural form 3",
                "the plural form 4",
                "the plural form 5",
            ]
        )
        out = BytesIO()
        store.serialize(out)

        assert out.getvalue().decode() == JSON_I18NEXT

    def test_nested_array(self) -> None:
        store = self.StoreClass()
        store.parse(JSON_I18NEXT_NESTED_ARRAY)

        assert len(store.units) == 4
        assert store.units[0].getid() == ".apps[0].title"
        assert store.units[1].getid() == ".apps[0].description"
        assert store.units[2].getid() == ".apps[1].title"
        assert store.units[3].getid() == ".apps[1].description"

        assert bytes(store).decode() == JSON_I18NEXT_NESTED_ARRAY

    def test_new_plural(self) -> None:
        EXPECTED = """{
    "simple": "the singular",
    "simple_plural": "the plural",
    "simple2": "the singular",
    "simple2_plural": "the plural",
    "complex_0": "the plural form 0",
    "complex_1": "the plural form 1",
    "complex_2": "the plural form 2",
    "complex_3": "the plural form 3",
    "complex_4": "the plural form 4",
    "complex_5": "the plural form 5"
}
"""
        store = self.StoreClass()
        store.settargetlanguage("ar")

        unit = self.StoreClass.UnitClass(
            multistring(
                [
                    "the singular",
                    "the plural",
                ]
            ),
            "simple",
        )
        store.addunit(unit)

        unit = self.StoreClass.UnitClass(
            multistring(
                [
                    "the singular",
                    "the plural",
                ]
            ),
        )
        unit.setid("simple2")
        store.addunit(unit)

        unit = self.StoreClass.UnitClass(
            "complex",
            "complex",
        )
        unit.target = multistring(
            [
                "the plural form 0",
                "the plural form 1",
                "the plural form 2",
                "the plural form 3",
                "the plural form 4",
                "the plural form 5",
                "the plural form 6",
            ]
        )
        store.addunit(unit)

        out = BytesIO()
        store.serialize(out)

        assert out.getvalue().decode() == EXPECTED

    def test_new_plural_id(self) -> None:
        EXPECTED = """{
    "simple_0": "the singular"
}
"""
        store = self.StoreClass()
        store.settargetlanguage("id")

        unit = self.StoreClass.UnitClass(
            multistring(
                [
                    "the singular",
                ]
            ),
            "simple",
        )
        store.addunit(unit)

        assert bytes(store).decode() == EXPECTED


class TestGoTextJsonFile(test_monolingual.TestMonolingualStore):
    StoreClass = jsonl10n.GoTextJsonFile

    def test_plurals(self) -> None:
        store = self.StoreClass()
        store.parse(JSON_GOTEXT)

        assert len(store.units) == 2
        assert store.units[0].target == multistring(["{N} tag", "{N} tags"])
        assert store.units[1].target == "Table"

        assert bytes(store).decode() == JSON_GOTEXT.decode()

    def test_plurals_missing(self) -> None:
        store = self.StoreClass()
        store.parse(JSON_GOTEXT)

        store.units[0].target = multistring(["{N} tag"])

        assert (
            '"other": {\n                            "msg": ""' in bytes(store).decode()
        )

    def test_case_no_msg(self) -> None:
        store = self.StoreClass()
        store.parse(
            """{
    "language": "en-US",
    "messages": [
        {
            "id": "{N} more files remaining!",
            "key": "%d more files remaining!",
            "message": "{N} more files remaining!",
            "translation": {
                "select": {
                    "feature": "plural",
                    "cases": {
                        "one": "One file remaining!",
                        "other": "There are {N} more files remaining!"
                    }
                }
            }
        }
    ]
}
"""
        )

        assert len(store.units) == 1
        assert store.units[0].target == multistring(
            ["One file remaining!", "There are {N} more files remaining!"]
        )

        assert (
            bytes(store).decode()
            == """{
    "language": "en-US",
    "messages": [
        {
            "id": "{N} more files remaining!",
            "message": "{N} more files remaining!",
            "key": "%d more files remaining!",
            "translation": {
                "select": {
                    "feature": "plural",
                    "cases": {
                        "one": {
                            "msg": "One file remaining!"
                        },
                        "other": {
                            "msg": "There are {N} more files remaining!"
                        }
                    }
                }
            }
        }
    ]
}
"""
        )

    def test_complex_id(self) -> None:
        text = """{
    "language": "en-US",
    "messages": [
        {
            "id": [
                "msgOutOfOrder",
                "{Device} is out of order!"
            ],
            "message": "{Device} is out of order!",
            "key": "%s is out of order!",
            "translation": ""
        },
        {
            "id": "[DEBUG] msg",
            "message": "[DEBUG] msg",
            "key": "[DEBUG] msg",
            "translation": ""
        }
    ]
}
"""
        store = self.StoreClass()
        store.parse(text)

        assert len(store.units) == 2
        assert (
            store.units[0].getid() == "['msgOutOfOrder', '{Device} is out of order!']"
        )
        assert store.units[1].getid() == "[DEBUG] msg"

        assert bytes(store).decode() == text


class TestI18NextV4Store(test_monolingual.TestMonolingualStore):
    StoreClass = jsonl10n.I18NextV4File
    DeepUnpluralizedJson = JSON_I18NEXT_PLURAL
    PartiallyPluralizedJson = JSON_I18NEXT_V4
    CorrectlyPluralizedJson = JSON_I18NEXT_V4_FIXED

    def test_serialize(self) -> None:
        store = self.StoreClass()
        store.targetlanguage = "ar"
        store.parse(JSON_I18NEXT_V4)
        out = BytesIO()
        store.serialize(out)

        # This will add missing plurals
        assert out.getvalue().decode() == JSON_I18NEXT_V4_FIXED

    def test_units(self) -> None:
        store = self.StoreClass()
        store.targetlanguage = "ar"
        store.parse(self.PartiallyPluralizedJson)
        assert len(store.units) == 4

    def test_plurals(self) -> None:
        store = self.StoreClass()
        store.targetlanguage = "ar"
        store.parse(self.PartiallyPluralizedJson)

        # Remove plurals
        store.units[2].target = "Ahoj"
        store.units[3].target = "Nazdar"
        out = BytesIO()
        store.serialize(out)

        assert out.getvalue() == self.DeepUnpluralizedJson

        # Bring back plurals
        store.settargetlanguage("ar")
        store.units[2].target = multistring(
            [
                "",
                "the singular",
                "",
                "",
                "",
                "the plural",
            ]
        )
        store.units[3].target = multistring(
            [
                "the plural form 0",
                "the plural form 1",
                "the plural form 2",
                "the plural form 3",
                "the plural form 4",
                "the plural form 5",
                "the plural form 6",
            ]
        )
        out = BytesIO()
        store.serialize(out)

        assert out.getvalue().decode() == self.CorrectlyPluralizedJson

    def test_nested_array(self) -> None:
        store = self.StoreClass()
        store.parse(JSON_I18NEXT_NESTED_ARRAY)

        assert len(store.units) == 4
        assert store.units[0].getid() == ".apps[0].title"
        assert store.units[1].getid() == ".apps[0].description"
        assert store.units[2].getid() == ".apps[1].title"
        assert store.units[3].getid() == ".apps[1].description"

        assert bytes(store).decode() == JSON_I18NEXT_NESTED_ARRAY

        newstore = self.StoreClass()
        newstore.parse(JSON_I18NEXT_NESTED_ARRAY_BEFORE)

        assert len(newstore.units) == 1

        newstore.addunit(store.units[0])
        newstore.addunit(store.units[1])
        newstore.addunit(store.units[2])
        newstore.addunit(store.units[3])

        assert bytes(newstore).decode() == JSON_I18NEXT_NESTED_ARRAY

    def test_new_plural(self) -> None:
        EXPECTED = """{
    "simple_zero": "the singular",
    "simple_one": "the plural",
    "complex_zero": "the plural form 0",
    "complex_one": "the plural form 1",
    "complex_two": "the plural form 2",
    "complex_few": "the plural form 3",
    "complex_many": "the plural form 4",
    "complex_other": "the plural form 5"
}
"""
        store = self.StoreClass()
        store.settargetlanguage("ar")

        unit = self.StoreClass.UnitClass(
            multistring(
                [
                    "the singular",
                    "the plural",
                ]
            ),
            "simple",
        )
        store.addunit(unit)

        unit = self.StoreClass.UnitClass(
            "complex",
            "complex",
        )
        store.addunit(unit)
        unit.target = multistring(
            [
                "the plural form 0",
                "the plural form 1",
                "the plural form 2",
                "the plural form 3",
                "the plural form 4",
                "the plural form 5",
            ]
        )

        assert bytes(store).decode() == EXPECTED

    def test_ru(self) -> None:
        data = """{
"bet_one": "еще +{{count}} ставка",
"bet_few": "еще +{{count}} ставки",
"bet_many": "еще +{{count}} ставок"
}
"""
        store = self.StoreClass()
        store.settargetlanguage("ru")
        store.parse(data)

        assert len(store.units) == 1
        assert store.units[0].source.strings == [
            "еще +{{count}} ставка",
            "еще +{{count}} ставки",
            "еще +{{count}} ставок",
        ]


class TestFlatI18NextV4Store(TestI18NextV4Store):
    StoreClass = jsonl10n.FlatI18NextV4File
    DeepUnpluralizedJson = JSON_FLAT_I18NEXT_V4_PLURAL
    PartiallyPluralizedJson = JSON_FLAT_I18NEXT_V4
    CorrectlyPluralizedJson = JSON_FLAT_I18NEXT_V4_FIXED


class TestGoI18NJsonFile(test_monolingual.TestMonolingualStore):
    StoreClass = jsonl10n.GoI18NJsonFile

    def test_plurals(self) -> None:
        store = self.StoreClass()
        store.parse(JSON_GOI18N)

        assert len(store.units) == 2
        assert store.units[0].target == multistring(
            ["{{.count}} tag", "{{.count}} tags"]
        )
        assert store.units[1].target == "Table"

        assert bytes(store).decode() == JSON_GOI18N.decode()

    def test_plurals_missing(self) -> None:
        store = self.StoreClass()
        store.parse(JSON_GOI18N)

        store.units[0].target = multistring(["{{.count}} tag"])

        assert '"other": ""' in bytes(store).decode()

    def test_invalid(self) -> None:
        store = self.StoreClass()
        with raises(ValueError):
            store.parse(JSON_I18NEXT_PLURAL)

    def test_dot_keys(self) -> None:
        jsontext = """[
    {
        "id": ".table",
        "translation": "message"
    }
]
"""
        store = self.StoreClass()
        store.parse(jsontext)
        # Edit target
        store.units[0].target = "message"
        assert bytes(store).decode() == jsontext


class TestGoI18NV2JsonFile(test_monolingual.TestMonolingualStore):
    StoreClass = jsonl10n.GoI18NV2JsonFile

    def test_plurals_1(self) -> None:
        store = self.StoreClass()
        store.parse(JSON_GOI18N_V2_SIMPLE)

        assert len(store.units) == 6
        assert store.units[0].target == "value"
        assert store.units[1].target == "Table"
        assert store.units[2].target == multistring(
            ["{{.count}} tag", "{{.count}} tags"]
        )
        assert store.units[3].target == "world"
        assert store.units[4].target == "hello"

        assert bytes(store).decode() == JSON_GOI18N_V2_SIMPLE

    def test_plurals_2(self) -> None:
        store = self.StoreClass()
        store.settargetlanguage("ar")
        store.parse(JSON_GOI18N_V2)

        # Remove plurals
        store.units[1].target = "plural"

        assert bytes(store).decode() == JSON_GOI18N_V2_PLURAL

        # Bring back plurals
        store.units[1].target = multistring(
            [
                "the plural form 0",
                "the plural form 1",
                "the plural form 2",
                "the plural form 3",
                "the plural form 4",
                "the plural form 5",
            ]
        )

        assert bytes(store).decode() == JSON_GOI18N_V2

    def test_plurals_blank(self) -> None:
        store = self.StoreClass()
        store.settargetlanguage("ar")
        store.parse(JSON_GOI18N_V2)
        # Remove one of plurals
        store.units[1].target = multistring(
            [
                "the plural form 0",
                "the plural form 1",
                "the plural form 2",
                "the plural form 3",
                "the plural form 4",
            ]
        )

        assert bytes(store).decode() == JSON_GOI18N_V2.replace("the plural form 5", "")

    def test_plurals_missing(self) -> None:
        store = self.StoreClass()
        store.parse(JSON_GOI18N_V2_SIMPLE)

        assert len(store.units) == 6

        store.units[2].target = multistring(["{{.count}} tag", "{{.count}} tags"])
        assert bytes(store).decode() == JSON_GOI18N_V2_SIMPLE

        store.units[2].target = multistring(["{{.count}} tag"])

        assert '"other": "{{.count}} tag"' in bytes(store).decode()

    def test_simplification(self) -> None:
        store = self.StoreClass()
        store.parse(JSON_GOI18N_V2_COMPLEX)

        assert len(store.units) == 6
        assert store.units[0].target == "value"
        assert store.units[1].target == "Table"
        assert store.units[2].target == multistring(
            ["{{.count}} tag", "{{.count}} tags"]
        )
        assert store.units[3].target == "world"
        assert store.units[4].target == "hello"

        assert bytes(store).decode() == JSON_GOI18N_V2_SIMPLE

    def test_invalid(self) -> None:
        store = self.StoreClass()
        with raises(ValueError):
            store.parse(JSON_I18NEXT_PLURAL)

    def test_dot_keys(self) -> None:
        jsontext = """{
    ".table": "message"
}
"""
        store = self.StoreClass()
        store.parse(jsontext)
        # Edit target
        store.units[0].target = "message"
        assert bytes(store).decode() == jsontext


class TestARBJsonFile(test_monolingual.TestMonolingualStore):
    StoreClass = jsonl10n.ARBJsonFile

    def test_roundtrip(self) -> None:
        store = self.StoreClass()
        store.parse(JSON_ARB)

        assert len(store.units) == 4
        assert store.units[0].isheader()
        assert store.units[1].target == "Back"
        assert store.units[2].target == "Next"
        assert store.units[3].target == "Done"

        assert bytes(store).decode() == JSON_ARB.decode()

    def test_leading_dot_keys(self) -> None:
        jsontext = """{
  ".dot": "dot",
  "@.dot": {
    "description": "description of .dot"
  },
  "nav:colon": "colon in a colon",
  "@nav:colon": {
    "description": "description of nav:colon"
  }
}
"""
        store = self.StoreClass()
        store.parse(jsontext)
        assert len(store.units) == 2
        store.units[1].target = "colon in a colon"
        assert bytes(store).decode() == jsontext

    def test_invalid_nesting(self) -> None:
        jsontext = """{
    "key": {
        ".dot": "dot",
        "@.dot": {}
    }
}
"""
        store = self.StoreClass()
        with raises(base.ParseError):
            store.parse(jsontext)

    def test_empty_metadata_not_stored(self) -> None:
        """Test that empty metadata is not stored in the output."""
        jsontext = """{
  "key1": "value1",
  "@key1": {},
  "key2": "value2",
  "@key2": {
    "type": "text"
  },
  "key3": "value3"
}
"""
        store = self.StoreClass()
        store.parse(jsontext)
        output = bytes(store).decode()

        assert '"@key1": {}' not in output

        # Non-empty metadata should be preserved
        assert '"@key2"' in output

        assert '"@key3"' not in output


JSON_FORMATJS = """{
    "hak27d": {
        "defaultMessage": "Control Panel",
        "description": "title of control panel section"
    },
    "haqsd": {
        "defaultMessage": "Delete user {name}",
        "description": "delete button"
    },
    "19hjs": {
        "defaultMessage": "New Password",
        "description": "placeholder text"
    },
    "explicit-id": {
        "defaultMessage": "Confirm Password",
        "description": "placeholder text"
    }
}
"""


class TestFormatJSJsonFile(test_monolingual.TestMonolingualStore):
    StoreClass = jsonl10n.FormatJSJsonFile

    def test_roundtrip(self) -> None:
        store = self.StoreClass()
        store.parse(JSON_FORMATJS)

        assert len(store.units) == 4
        assert store.units[3].getid() == "explicit-id"
        assert store.units[3].target == "Confirm Password"
        assert store.units[3].getnotes() == "placeholder text"

        assert bytes(store).decode() == JSON_FORMATJS

    def test_leading_dot_keys(self) -> None:
        jsontext = """{
    ".dot": {
        "defaultMessage": "Control Panel",
        "description": "title of control panel section"
    }
}
"""
        store = self.StoreClass()
        store.parse(jsontext)
        assert bytes(store).decode() == jsontext

    def test_invalid(self) -> None:
        jsontext = """{
    ".dot": "text"
}
"""
        store = self.StoreClass()
        with raises(base.ParseError):
            store.parse(jsontext)


JSON_NEXTCLOUD_SIMPLE = b"""{
    "translations": {
        "Hello": "Hallo",
        "Goodbye": "Auf Wiedersehen"
    }
}
"""

JSON_NEXTCLOUD_WITH_PLURAL_FORM = rb"""{
    "translations": {
        "Hello": "Hallo",
        "_%n tree_::_%n trees_": [
            "%n Baum",
            "%n B\u00e4ume"
        ]
    },
    "pluralForm": "nplurals=2; plural=(n != 1);"
}
"""

# spellchecker:off
JSON_NEXTCLOUD_COMPLEX = rb"""{
    "translations": {
        "Simple string": "Einfache Zeichenkette",
        "_%n comment_::_%n comments_": [
            "%n Kommentar",
            "%n Kommentare"
        ],
        "Another key": "Ein anderer Wert"
    },
    "pluralForm": "nplurals=2; plural=(n != 1);"
}
"""
# spellchecker:on


class TestNextcloudJsonUnit(test_monolingual.TestMonolingualUnit):
    UnitClass = jsonl10n.NextcloudJsonUnit

    def test_source_property_maps_to_id(self) -> None:
        """Test that source property maps to unit ID for NextcloudJsonUnit."""
        unit = self.UnitClass("target value")
        unit.setid("my_key")
        assert unit.source == "my_key"

        # Setting source should update the ID
        unit.source = "new_key"
        assert unit.getid() == "new_key"
        assert unit.source == "new_key"

    def test_source_roundtrip(self) -> None:
        """Test that source property persists through get/set."""
        unit = self.UnitClass("Hello World")
        unit.source = "greeting"
        assert unit.getid() == "greeting"
        assert unit.source == "greeting"


class TestNextcloudJsonFile(test_monolingual.TestMonolingualStore):
    StoreClass = jsonl10n.NextcloudJsonFile

    def test_parse_simple(self) -> None:
        """Test parsing simple Nextcloud JSON with no plurals."""
        store = self.StoreClass()
        store.parse(JSON_NEXTCLOUD_SIMPLE)

        assert len(store.units) == 2
        assert store.units[0].getid() == "Hello"
        assert store.units[0].target == "Hallo"
        assert store.units[1].getid() == "Goodbye"
        assert store.units[1].target == "Auf Wiedersehen"

    def test_serialize_simple(self) -> None:
        """Test serializing simple translations."""
        store = self.StoreClass()
        store.parse(JSON_NEXTCLOUD_SIMPLE)

        out = BytesIO()
        store.serialize(out)
        result = out.getvalue()

        assert b'"translations"' in result
        assert b'"Hello": "Hallo"' in result
        assert b'"Goodbye": "Auf Wiedersehen"' in result

    def test_parse_with_plurals(self) -> None:
        """Test parsing Nextcloud JSON with plural forms."""
        store = self.StoreClass()
        store.parse(JSON_NEXTCLOUD_WITH_PLURAL_FORM)

        assert len(store.units) == 2

        # First unit should be simple string
        assert store.units[0].getid() == "Hello"
        assert store.units[0].target == "Hallo"
        assert not isinstance(store.units[0].target, multistring)

        # Second unit should be plural
        assert store.units[1].getid() == "_%n tree_::_%n trees_"
        assert isinstance(store.units[1].target, multistring)
        assert len(store.units[1].target.strings) == 2
        assert store.units[1].target.strings[0] == "%n Baum"
        assert store.units[1].target.strings[1] == "%n Bäume"

    def test_preserve_plural_form(self) -> None:
        """Test that pluralForm metadata is preserved during round-trip."""
        store = self.StoreClass()
        store.parse(JSON_NEXTCLOUD_WITH_PLURAL_FORM)

        out = BytesIO()
        store.serialize(out)
        result = out.getvalue().decode()

        # Verify pluralForm is preserved
        assert '"pluralForm": "nplurals=2; plural=(n != 1);"' in result
        assert '"translations"' in result

    def test_roundtrip_with_plurals(self) -> None:
        """Test full round-trip with plurals and metadata."""
        store = self.StoreClass()
        store.parse(JSON_NEXTCLOUD_COMPLEX)

        assert len(store.units) == 3

        # Verify parsing
        assert store.units[0].target == "Einfache Zeichenkette"
        assert isinstance(store.units[1].target, multistring)
        assert store.units[2].target == "Ein anderer Wert"  # codespell:ignore

        # Serialize and re-parse
        out = BytesIO()
        store.serialize(out)

        store2 = self.StoreClass()
        store2.parse(out.getvalue())

        assert len(store2.units) == 3
        assert store2.units[0].target == "Einfache Zeichenkette"
        assert isinstance(store2.units[1].target, multistring)
        assert len(store2.units[1].target.strings) == 2
        assert store2.units[2].target == "Ein anderer Wert"  # codespell:ignore

    def test_ignore_non_translations_keys(self) -> None:
        """Test that only 'translations' key is parsed for units."""
        json_with_extra = b"""{
    "someOtherKey": "should be ignored",
    "translations": {
        "Hello": "Hallo"
    },
    "pluralForm": "nplurals=2; plural=(n != 1);"
}
"""
        store = self.StoreClass()
        store.parse(json_with_extra)

        # Should only have 1 unit from translations
        assert len(store.units) == 1
        assert store.units[0].getid() == "Hello"

    def test_preserve_other_metadata(self) -> None:
        """Test that arbitrary metadata outside translations is preserved."""
        json_with_metadata = b"""{
    "translations": {
        "Hello": "Hallo"
    },
    "pluralForm": "nplurals=2; plural=(n != 1);",
    "customField": "customValue"
}
"""
        store = self.StoreClass()
        store.parse(json_with_metadata)

        out = BytesIO()
        store.serialize(out)
        result = out.getvalue().decode()

        # Verify all metadata is preserved
        assert '"pluralForm": "nplurals=2; plural=(n != 1);"' in result
        assert '"customField": "customValue"' in result
        assert '"translations"' in result

    def test_add_unit(self) -> None:
        """Test adding a new unit to Nextcloud JSON."""
        store = self.StoreClass()
        store.parse(JSON_NEXTCLOUD_SIMPLE)

        # Add a new simple unit
        unit = self.StoreClass.UnitClass("Neuer Wert")
        unit.setid("New Key")
        store.addunit(unit)

        assert len(store.units) == 3
        assert store.units[2].getid() == "New Key"
        assert store.units[2].target == "Neuer Wert"

    def test_add_plural_unit(self) -> None:
        """Test adding a plural unit to Nextcloud JSON."""
        store = self.StoreClass()
        store.parse(JSON_NEXTCLOUD_SIMPLE)

        # Add a plural unit
        unit = self.StoreClass.UnitClass(multistring(["%n Datei", "%n Dateien"]))
        unit.setid("_%n file_::_%n files_")
        store.addunit(unit)

        out = BytesIO()
        store.serialize(out)
        result = out.getvalue().decode()

        # Verify plural is serialized as array
        assert (
            '"_%n file_::_%n files_": [\n            "%n Datei",\n            "%n Dateien"\n        ]'
            in result
        )

    def test_empty_translations(self) -> None:
        """Test handling of empty translations object."""
        json_empty = b"""{
    "translations": {},
    "pluralForm": "nplurals=2; plural=(n != 1);"
}
"""
        store = self.StoreClass()
        store.parse(json_empty)

        assert len(store.units) == 0

        out = BytesIO()
        store.serialize(out)
        result = out.getvalue().decode()

        assert '"translations": {}' in result
        assert '"pluralForm"' in result


JSON_RESJSON = b"""{
    "greeting": "Hello",
    "_greeting.comment": "A welcome greeting.",
    "_greeting.source": "Hello",
    "farewell": "Goodbye",
    "_farewell.comment": "A parting message.",
    "_farewell.source": "Goodbye"
}
"""


class TestRESJSONFile(test_monolingual.TestMonolingualStore):
    StoreClass = jsonl10n.RESJSONFile

    def test_roundtrip(self) -> None:
        store = self.StoreClass()
        store.parse(JSON_RESJSON)

        assert len(store.units) == 2
        assert store.units[0].target == "Hello"
        assert store.units[0].getnotes() == "A welcome greeting."
        assert store.units[0].metadata.get("source") == "Hello"
        assert store.units[1].target == "Goodbye"
        assert store.units[1].getnotes() == "A parting message."
        assert store.units[1].metadata.get("source") == "Goodbye"

        assert bytes(store).decode() == JSON_RESJSON.decode()

    def test_basic_parsing(self) -> None:
        jsontext = """{
    "key": "value",
    "_key.comment": "A comment"
}
"""
        store = self.StoreClass()
        store.parse(jsontext)
        assert len(store.units) == 1
        assert store.units[0].target == "value"
        assert store.units[0].getnotes() == "A comment"
        assert bytes(store).decode() == jsontext

    def test_multiple_metadata(self) -> None:
        jsontext = """{
    "message": "text",
    "_message.comment": "comment text",
    "_message.source": "source text",
    "_message.custom": "custom data"
}
"""
        store = self.StoreClass()
        store.parse(jsontext)
        assert len(store.units) == 1
        assert store.units[0].target == "text"
        assert store.units[0].getnotes() == "comment text"
        assert store.units[0].metadata.get("source") == "source text"
        assert store.units[0].metadata.get("custom") == "custom data"
        assert bytes(store).decode() == jsontext

    def test_no_metadata(self) -> None:
        jsontext = """{
    "simple": "value"
}
"""
        store = self.StoreClass()
        store.parse(jsontext)
        assert len(store.units) == 1
        assert store.units[0].target == "value"
        assert store.units[0].getnotes() == ""
        assert bytes(store).decode() == jsontext

    def test_edit_target(self) -> None:
        jsontext = """{
    "key": "value",
    "_key.comment": "comment"
}
"""
        store = self.StoreClass()
        store.parse(jsontext)
        store.units[0].target = "new value"
        result = bytes(store).decode()
        assert '"key": "new value"' in result
        assert '"_key.comment": "comment"' in result

    def test_edit_notes(self) -> None:
        jsontext = """{
    "key": "value",
    "_key.comment": "original comment",
    "_key.source": "source text"
}
"""
        store = self.StoreClass()
        store.parse(jsontext)
        store.units[0].notes = "modified comment"
        result = bytes(store).decode()
        assert '"key": "value"' in result
        assert '"_key.comment": "modified comment"' in result
        assert '"_key.source": "source text"' in result

    def test_keys_with_dots(self) -> None:
        jsontext = """{
    "foo.bar": "value",
    "_foo.bar.comment": "comment for foo.bar"
}
"""
        store = self.StoreClass()
        store.parse(jsontext)
        assert len(store.units) == 1
        assert store.units[0].getid() == "foo.bar"
        assert store.units[0].target == "value"
        assert store.units[0].getnotes() == "comment for foo.bar"
        assert bytes(store).decode() == jsontext

    def test_leading_dot_keys(self) -> None:
        jsontext = """{
    ".dot": "dot value",
    "_.dot.comment": "comment"
}
"""
        store = self.StoreClass()
        store.parse(jsontext)
        assert len(store.units) == 1
        assert store.units[0].target == "dot value"
        assert bytes(store).decode() == jsontext

    def test_invalid_nesting(self) -> None:
        jsontext = """{
    "key": {
        "nested": "value"
    }
}
"""
        store = self.StoreClass()
        with raises(base.ParseError):
            store.parse(jsontext)

    def test_source_property_get_set(self) -> None:
        """Test that source property can be read and written correctly."""
        jsontext = """{
    "key": "value",
    "_key.source": "original source"
}
"""
        store = self.StoreClass()
        store.parse(jsontext)
        assert len(store.units) == 1
        assert store.units[0].source == "original source"
        assert store.units[0].target == "value"

        # Modify source
        store.units[0].source = "modified source"
        result = bytes(store).decode()
        assert '"_key.source": "modified source"' in result

    def test_source_property_persists(self) -> None:
        """Test that source property persists through serialization."""
        store = self.StoreClass()
        unit = store.UnitClass("")
        unit.setid("mykey")
        unit.target = "target value"
        unit.source = "source value"
        store.addunit(unit)

        result = bytes(store).decode()
        assert '"mykey": "target value"' in result
        assert '"_mykey.source": "source value"' in result

        # Parse again to verify persistence
        store2 = self.StoreClass()
        store2.parse(result)
        assert len(store2.units) == 1
        assert store2.units[0].source == "source value"
        assert store2.units[0].target == "target value"

    def test_getcontext_returns_id(self) -> None:
        """Test that getcontext() returns the unit ID."""
        jsontext = """{
    "greeting": "Hello",
    "_greeting.comment": "A greeting"
}
"""
        store = self.StoreClass()
        store.parse(jsontext)
        assert len(store.units) == 1
        assert store.units[0].getcontext() == "greeting"

    def test_parsing_preserves_order(self) -> None:
        """Test that parsing preserves the order of keys."""
        jsontext = """{
    "first": "1",
    "_first.source": "one",
    "second": "2",
    "_second.source": "two",
    "third": "3",
    "_third.source": "three"
}
"""
        store = self.StoreClass()
        store.parse(jsontext)
        assert len(store.units) == 3
        assert store.units[0].getid() == "first"
        assert store.units[1].getid() == "second"
        assert store.units[2].getid() == "third"

    def test_metadata_without_translation(self) -> None:
        """Test handling when metadata exists but no corresponding translation value."""
        jsontext = """{
    "key": "value",
    "_key.source": "source text",
    "_orphan.comment": "orphan metadata"
}
"""
        store = self.StoreClass()
        store.parse(jsontext)
        # Should handle orphan metadata gracefully
        # The parsing code extracts "orphan" as the base key from "_orphan.comment"
        assert len(store.units) == 2
        # First unit should be the normal key with metadata
        assert store.units[0].getid() == "key"
        assert store.units[0].source == "source text"
        # Second unit should extract the base key from the orphan metadata
        assert store.units[1].getid() == "orphan"
        assert store.units[1].metadata.get("comment") == "orphan metadata"

    def test_complex_keys_with_multiple_dots(self) -> None:
        """Test handling of keys with multiple dots."""
        jsontext = """{
    "app.section.key": "value",
    "_app.section.key.source": "source",
    "_app.section.key.comment": "comment"
}
"""
        store = self.StoreClass()
        store.parse(jsontext)
        assert len(store.units) == 1
        assert store.units[0].getid() == "app.section.key"
        assert store.units[0].target == "value"
        assert store.units[0].source == "source"
        assert store.units[0].getnotes() == "comment"

    def test_source_empty_string(self) -> None:
        """Test that empty source strings are handled correctly."""
        jsontext = """{
    "key": "value",
    "_key.source": ""
}
"""
        store = self.StoreClass()
        store.parse(jsontext)
        assert len(store.units) == 1
        assert store.units[0].source == ""
        assert store.units[0].target == "value"



import os

from translate.convert import po2odf

from . import test_convert


class TestPO2ODFCommand(test_convert.TestConvertCommand):
    """Tests running actual po2odf commands on files."""

    convertmodule = po2odf
    expected_options = [
        "-t TEMPLATE, --template=TEMPLATE",
    ]

    def test_convert(self) -> None:
        """Test basic PO to ODF conversion."""
        posource = """# Translation file
msgid ""
msgstr ""
"Content-Type: text/plain; charset=UTF-8\\n"

msgid "Hello, world!"
msgstr "Ahoj světe!"
"""
        self.create_testfile("simple.po", posource)
        self.run_command(
            i="simple.po",
            o="simple.odt",
            template=os.path.join(os.path.dirname(__file__), "test.odt"),
        )
        # Check that the output file was created
        assert os.path.exists(self.get_testfilename("simple.odt"))

    def test_convert_with_multiple_units(self) -> None:
        """Test PO to ODF conversion with multiple translation units."""
        posource = """# Translation file
msgid ""
msgstr ""
"Content-Type: text/plain; charset=UTF-8\\n"

msgid "First"
msgstr "První"

msgid "Second"
msgstr "Druhý"

msgid "Third"
msgstr "Třetí"
"""
        self.create_testfile("multi.po", posource)
        self.run_command(
            i="multi.po",
            o="multi.odt",
            template=os.path.join(os.path.dirname(__file__), "test.odt"),
        )
        # Check that the output file was created
        assert os.path.exists(self.get_testfilename("multi.odt"))



#
# Copyright 2002-2014 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

from lxml import etree

from translate.storage.xml_extract import extract, misc
from translate.storage.xml_name import XmlNamer


def _get_tag_arrays(dom_node):
    """
    Return a dictionary indexed by child tag names, where each tag is associated with an array
    of all the child nodes with matching the tag name, in the order in which they appear as children
    of dom_node.

    >>> xml = '<a><b></b><c></c><b></b><d/></a>'
    >>> element = etree.fromstring(xml)
    >>> get_tag_arrays(element)
    {'b': [<Element a at 84df144>, <Element a at 84df148>], 'c': [<Element a at 84df120>], 'd': [<Element a at 84df152>]}
    """
    child_dict = {}
    for child in dom_node:
        if child.tag not in child_dict:
            child_dict[child.tag] = []
        child_dict[child.tag].append(child)
    return child_dict


def apply_translations(dom_node, unit_node, do_translate) -> None:
    tag_array = _get_tag_arrays(dom_node)
    for unit_child_index, unit_child in unit_node.children.items():
        tag, index = unit_child_index
        try:
            dom_child = tag_array[XmlNamer(dom_node).name(tag)][index]
            apply_translations(dom_child, unit_child, do_translate)
        # Raised if tag is not in tag_array. We might want to complain to the
        # user in the future.
        except KeyError:
            pass
        # Raised if index is not in tag_array[tag]. We might want to complain to
        # the user in the future
        except IndexError:
            pass
    # If there is a translation unit associated with this unit_node...
    if unit_node.unit is not None:
        # The invoke do_translate on the dom_node and the unit; do_translate
        # should replace the text in dom_node with the text in unit_node.
        do_translate(dom_node, unit_node.unit)


def reduce_dom_tree(f, dom_node, *state):
    return misc.reduce_tree(f, dom_node, dom_node, lambda dom_node: dom_node, *state)


def find_dom_root(parent_dom_node, dom_node):
    """.. seealso:: :meth:`find_placeable_dom_tree_roots`."""
    if dom_node is None or parent_dom_node is None:
        return None
    if dom_node.getparent() == parent_dom_node:
        return dom_node
    if dom_node.getparent() is None:
        return None
    return find_dom_root(parent_dom_node, dom_node.getparent())


def find_placeable_dom_tree_roots(unit_node):
    r"""
    For an inline placeable, find the root DOM node for the placeable in its
    parent.

    Consider the diagram. In this pseudo-ODF example, there is an inline span
    element. However, the span is contained in other tags (which we never process).
    When splicing the template DOM tree (that is, the DOM which comes from
    the XML document we're using to generate a translated XML document), we'll
    need to move DOM sub-trees around and we need the roots of these sub-trees::

        <p> This is text \/                <- Paragraph containing an inline placeable
                         <blah>            <- Inline placeable's root (which we want to find)
                         ...               <- Any number of intermediate DOM nodes
                         <span> bold text  <- The inline placeable's Translatable
                                              holds a reference to this DOM node
    """

    def set_dom_root_for_unit_node(parent_unit_node, unit_node, dom_tree_roots):
        dom_tree_roots[unit_node] = find_dom_root(
            parent_unit_node.dom_node, unit_node.dom_node
        )
        return dom_tree_roots

    return extract.reduce_unit_tree(set_dom_root_for_unit_node, unit_node, {})


def _map_source_dom_to_doc_dom(unit_node, source_dom_node):
    r"""
    Creating a mapping from the DOM nodes in source_dom_node which correspond to
    placeables, with DOM nodes in the XML document template (this information is obtained
    from unit_node). We are interested in DOM nodes in the XML document template which
    are the roots of placeables. See the diagram below, as well as
    :meth:`find_placeable_dom_tree_roots`.

    XLIFF Source (below)::
        <source>This is text <g> bold text</g> and a footnote<x/></source>
                             /                                 \________
                            /                                           \
        <p>This is text<blah>...<span> bold text</span>...</blah> and <note>...</note></p>
    Input XML document used as a template (above)

    In the above diagram, the XLIFF source DOM node <g> is associated with the XML
    document DOM node <blah>, whereas the XLIFF source DOM node <x> is associated with
    the XML document DOM node <note>.
    """
    dom_tree_roots = find_placeable_dom_tree_roots(unit_node)
    source_dom_to_doc_dom = {}

    def loop(unit_node, source_dom_node) -> None:
        for child_unit_node, child_source_dom in zip(
            unit_node.placeables, source_dom_node, strict=False
        ):
            source_dom_to_doc_dom[child_source_dom] = dom_tree_roots[child_unit_node]
            loop(child_unit_node, child_source_dom)

    loop(unit_node, source_dom_node)
    return source_dom_to_doc_dom


def _map_target_dom_to_source_dom(source_dom_node, target_dom_node):
    """
    Associate placeables in source_dom_node and target_dom_node which
    have the same 'id' attributes.

    We're using XLIFF placeables. The XLIFF standard requires that
    placeables have unique ids. The id of a placeable is never modified,
    which means that even if placeables are moved around in a translation,
    we can easily associate placeables from the source text with placeables
    in the target text.

    This function does exactly that.
    """

    def map_id_to_dom_node(parent_node, node, id_to_dom_node):
        # If this DOM node has an 'id' attribute, then add an id -> node
        # mapping to 'id_to_dom_node'.
        if "id" in node.attrib:
            id_to_dom_node[node.attrib["id"]] = node
        return id_to_dom_node

    # Build a mapping of id attributes to the DOM nodes which have these ids.
    id_to_dom_node = reduce_dom_tree(map_id_to_dom_node, target_dom_node, {})

    def map_target_dom_to_source_dom_aux(parent_node, node, target_dom_to_source_dom):
        if "id" in node.attrib and node.attrib["id"] in id_to_dom_node:
            target_dom_to_source_dom[id_to_dom_node[node.attrib["id"]]] = node
        return target_dom_to_source_dom

    # For each node in the DOM tree rooted at source_dom_node:
    # 1. Check whether the node has an 'id' attribute.
    # 2. If so, check whether there is a mapping of this id to a target DOM node
    #    in id_to_dom_node.
    # 3. If so, associate this source DOM node with the target DOM node.
    return reduce_dom_tree(map_target_dom_to_source_dom_aux, source_dom_node, {})


def _build_target_dom_to_doc_dom(unit_node, source_dom, target_dom):
    source_dom_to_doc_dom = _map_source_dom_to_doc_dom(unit_node, source_dom)
    target_dom_to_source_dom = _map_target_dom_to_source_dom(source_dom, target_dom)
    return misc.compose_mappings(target_dom_to_source_dom, source_dom_to_doc_dom)


def _get_translated_node(target_node, target_dom_to_doc_dom):
    """
    Convenience function to get node corresponding to 'target_node' and to
    assign the tail text of 'target_node' to this node.
    """
    dom_node = target_dom_to_doc_dom[target_node]
    dom_node.tail = target_node.tail
    return dom_node


def _build_translated_dom(dom_node, target_node, target_dom_to_doc_dom) -> None:
    """
    Use the "shape" of 'target_node' (which is a DOM tree) to insert nodes
    into the DOM tree rooted at 'dom_node'.

    The mapping 'target_dom_to_doc_dom' is used to map nodes from 'target_node'
    to nodes which much be inserted into dom_node.
    """
    if isinstance(dom_node, etree._ProcessingInstruction):
        # If the DOM node is a processing instruction we can safely skip it,
        # because we don't want to change it and it has no children.
        return

    dom_node.text = target_node.text
    # 1. Find all child nodes of target_node.
    # 2. Filter out the children which map to None.
    # 3. Call _get_translated_node on the remaining children; this maps a node
    #    in 'target_node' to a node in 'dom_node' and assigns the tail text of
    #    'target_node' to the mapped node.
    # 4. Add all of these mapped nodes to 'dom_node'
    dom_node.extend(
        _get_translated_node(child, target_dom_to_doc_dom)
        for child in target_node
        if target_dom_to_doc_dom[child] is not None
    )
    # Recursively call this function on pairs of matched children in
    # dom_node and target_node.
    for dom_child, target_child in zip(dom_node, target_node, strict=False):
        _build_translated_dom(dom_child, target_child, target_dom_to_doc_dom)


def get_xliff_source_target_doms(unit):
    """
    Return a tuple with unit source and target DOM objects.

    This method is method is meant to provide a way to retrieve the DOM objects
    for the unit source and target for XLIFF stores.
    """
    if unit.target_dom is not None:
        return (unit.source_dom, unit.target_dom)

    return (unit.source_dom, unit.source_dom)


def replace_dom_text(
    make_parse_state,
    dom_retriever=get_xliff_source_target_doms,
    process_translatable=extract.process_translatable,
):
    """
    Generate a replacement functions.

    Return a function::

          action: etree_Element x base.TranslationUnit -> None

    which takes a dom_node and a translation unit. The dom_node is rearranged
    according to rearrangement of placeables in unit.target (relative to their
    positions in unit.source).
    """

    def action(dom_node, unit) -> None:
        """
        Use the unit's target (or source in the case where there is no
        translation) to update the text in the dom_node and at the tails of its
        children.
        """
        source_dom, target_dom = dom_retriever(unit)

        # Build a tree of (non-DOM) nodes which correspond to the translatable DOM nodes in 'dom_node'.
        # Pass in a fresh parse_state every time, so as avoid working with stale parse state info.
        unit_node = extract.find_translatable_dom_nodes(
            dom_node, make_parse_state(), process_translatable
        )[0]
        target_dom_to_doc_dom = _build_target_dom_to_doc_dom(
            unit_node, source_dom, target_dom
        )
        # Before we start reconstructing the sub-tree rooted at dom_node, we must clear out its children
        dom_node[:] = []
        _build_translated_dom(dom_node, target_dom, target_dom_to_doc_dom)

    return action
